from functools import partial
from itertools import accumulate, product
from numbers import Number

import numpy as np

from ..core.utils import cached_property
from .abstract import CompressedArray
from .mixin import CompressedArrayMixin
from .netcdfindexer import netcdf_indexer
from .subarray import (
    BiLinearSubarray,
    BiQuadraticLatitudeLongitudeSubarray,
    InterpolationSubarray,
    LinearSubarray,
    QuadraticLatitudeLongitudeSubarray,
    QuadraticSubarray,
)


class SubsampledArray(CompressedArrayMixin, CompressedArray):
    """An underlying subsampled array.

    For some structured coordinate data (e.g. coordinates describing
    remote sensing products) space may be saved by storing a subsample
    of the data, called tie points. The uncompressed data can be
    reconstituted by interpolation, from the subsampled values. This
    process will likely result in a loss in accuracy (as opposed to
    precision) in the uncompressed variables, due to rounding and
    approximation errors in the interpolation calculations, but it is
    assumed that these errors will be small enough to not be of
    concern to users of the uncompressed dataset. The creator of the
    compressed dataset can control the accuracy of the reconstituted
    data through the degree of subsampling and the choice of
    interpolation method.

    See CF section 8.3 "Lossy Compression by Coordinate Subsampling"
    and Appendix J "Coordinate Interpolation Methods".

    >>> tie_point_indices={{package}}.TiePointIndex(data=[0, 4, 7, 8, 11])
    >>> w = {{package}}.InterpolationParameter(data=[5, 10, 5])
    >>> coords = {{package}}.SubsampledArray(
    ...     interpolation_name='quadratic',
    ...     compressed_array={{package}}.Data([15, 135, 225, 255, 345]),
    ...     shape=(12,),
    ...     tie_point_indices={0: tie_point_indices},
    ...     parameters={"w": w},
    ...     parameter_dimensions={"w": (0,)},
    ... )
    >>> print(coords[...])
    [ 15.          48.75        80.         108.75       135.
     173.88888889 203.88888889 225.         255.         289.44444444
     319.44444444 345.        ]

    **Cell boundaries**

    When the tie points array represents bounds tie points then the
    *shape* parameter describes the uncompressed bounds shape. See CF
    section 8.3.9 "Interpolation of Cell Boundaries".

    >>> bounds = {{package}}.SubsampledArray(
    ...     interpolation_name='quadratic',
    ...     compressed_array={{package}}.Data([0, 150, 240, 240, 360]),
    ...     shape=(12, 2),
    ...     tie_point_indices={0: tie_point_indices},
    ...     parameters={"w": w},
    ...     parameter_dimensions={"w": (0,)},
    ... )
    >>> print(bounds[...])
    [[0.0 33.2]
     [33.2 64.8]
     [64.8 94.80000000000001]
     [94.80000000000001 123.2]
     [123.2 150.0]
     [150.0 188.88888888888889]
     [188.88888888888889 218.88888888888889]
     [218.88888888888889 240.0]
     [240.0 273.75]
     [273.75 305.0]
     [305.0 333.75]
     [333.75 360.0]]

    .. versionadded:: (cfdm) 1.10.0.0

    """

    _flag_names = ("location_use_3d_cartesian",)

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        instance = super().__new__(cls)

        instance._Subarray = {
            "bi_linear": BiLinearSubarray,
            "bi_quadratic_latitude_longitude": BiQuadraticLatitudeLongitudeSubarray,
            "linear": LinearSubarray,
            "quadratic": QuadraticSubarray,
            "quadratic_latitude_longitude": QuadraticLatitudeLongitudeSubarray,
            None: InterpolationSubarray,
        }

        return instance

    def __init__(
        self,
        interpolation_name=None,
        compressed_array=None,
        shape=None,
        computational_precision=None,
        interpolation_description=None,
        tie_point_indices=None,
        parameters=None,
        parameter_dimensions=None,
        dependent_tie_points=None,
        dependent_tie_point_dimensions=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            interpolation_name: `str`, optional
                The name of the interpolation method.

                Standardised interpolation method are defined in CF
                appendix J "Coordinate Interpolation Methods".

                If not set then the non-standardised interpolation
                method is assumed to be defined by the
                *interpolation_description* parameter

                *Parameter example:*
                  ``'bi_linear'``

            compressed_array: array_like
                The (bounds) tie points array.

            shape: `tuple`
                The uncompressed array shape.

            tie_point_indices: `dict`
                The tie point index variable for each subsampled
                dimension. A key identifies a subsampled dimension by
                its integer position in the compressed array, and its
                value is a `TiePointIndex` variable.

                *Parameter example:*
                  ``{1: {{package}}.TiePointIndex(data=[0, 16, 31])}``

            computational_precision: `str`, optional
                The floating-point arithmetic precision used during
                the preparation and validation of the compressed data.

                *Parameter example:*
                  ``'64'``

            interpolation_description: `str`, optional
                A complete non-standardised description of the
                interpolation method.

            parameters: `dict`, optional
                Interpolation parameters required by the interpolation
                method. Each key is an interpolation parameter term
                name, whose value is an `InterpolationParameter`
                variable.

                Interpolation parameter term names for the
                standardised interpolation methods are defined in CF
                Appendix J "Coordinate Interpolation Methods".

                Set to an empty dictionary for interpolation methods
                that do not require interpolation parameters.

                *Parameter example:*
                  ``{}``

                *Parameter example:*
                  ``{'w': {{package}}.InterpolationParameter(data=[5, 10, 5])}``

            parameter_dimensions: `dict`, optional
                The tie point array dimensions that correspond to the
                dimensions of the interpolation parameters, keyed by the
                interpolation parameter term names.

                Interpolation parameter term names for the
                standardised interpolation methods are defined in CF
                Appendix J "Coordinate Interpolation Methods".

                Set to an empty dictionary for interpolation methods
                that do not require interpolation parameters.

                *Parameter example:*
                  ``{}``

                *Parameter example:*
                  ``{'w': (0,)}``

                *Parameter example:*
                  ``{'ce1': (1, 2), 'ca2': (2, 1)}``

            dependent_tie_points: `dict`, optional
                The dependent tie point arrays needed by the
                interpolation method, keyed by the dependent tie point
                identities. Each key is a dependent tie point
                identity, whose value is a `Data` variable.

                The identities must be understood the interpolation
                method. When *interpolation_name* is
                ``'quadratic_latitude_longitude'`` or
                ``'bi_quadratic_latitude_longitude'` then the
                *dependent_tie_points* dictionary must contain exactly
                one of the keys ``'latitude'`` or ``'longitude``.

                Set to an empty dictionary for interpolation methods
                that do not require dependent tie points.

                *Parameter example:*
                  ``{}``

                *Parameter example:*
                  ``{'latitude': {{package}}.Data([30, 31, 32])}``

            dependent_tie_points_dimensions: `dict`, optional
                The tie point array dimensions that correspond to the
                dimensions of the dependent tie points arrays, keyed
                the dependent tie point identities.

                Set to an empty dictionary for interpolation methods
                that do not require dependent tie points.

                *Parameter example:*
                  ``{}``

                *Parameter example:*
                  ``{'latitude': (0,)}``

                *Parameter example:*
                  ``{'longitude': (0, 1)}``

                *Parameter example:*
                  ``{'latitude': (2, 0, 1)}``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            compressed_array=compressed_array,
            shape=shape,
            compression_type="subsampled",
            source=source,
            copy=copy,
        )

        compressed_dimensions = None

        if source is not None:
            try:
                interpolation_name = source.get_interpolation_name(None)
            except AttributeError:
                interpolation_name = None

            try:
                computational_precision = source.get_computational_precision(
                    None
                )
            except AttributeError:
                computational_precision = None

            try:
                interpolation_description = (
                    source.get_interpolation_description(None)
                )
            except AttributeError:
                interpolation_description = None

            try:
                tie_point_indices = source.get_tie_point_indices()
            except AttributeError:
                tie_point_indices = {}

            try:
                parameters = source.get_parameters()
            except AttributeError:
                parameters = {}

            try:
                parameter_dimensions = source.get_parameter_dimensions()
            except AttributeError:
                parameter_dimensions = {}

            try:
                dependent_tie_points = source.get_dependent_tie_points()
            except AttributeError:
                dependent_tie_points = {}

            try:
                dependent_tie_point_dimensions = (
                    source.get_dependent_tie_point_dimensions()
                )
            except AttributeError:
                dependent_tie_point_dimensions = {}

            try:
                compressed_dimensions = source.compressed_dimensions()
            except AttributeError:
                compressed_dimensions = {}
        else:
            if tie_point_indices is None:
                tie_point_indices = {}

            if parameters is None:
                parameters = {}

            if parameter_dimensions is None:
                parameter_dimensions = {}

            if dependent_tie_points is None:
                dependent_tie_points = {}

            if dependent_tie_point_dimensions is None:
                dependent_tie_point_dimensions = {}

        if compressed_dimensions is None:
            compressed_dimensions = {
                d: (d,) for d in sorted(tie_point_indices)
            }

        self._set_component(
            "compressed_dimensions", compressed_dimensions, copy=False
        )

        if interpolation_name is not None:
            self._set_component(
                "interpolation_name", interpolation_name, copy=False
            )

        if interpolation_description is not None:
            self._set_component(
                "interpolation_description",
                interpolation_description,
                copy=False,
            )

        if computational_precision is not None:
            self._set_component(
                "computational_precision", computational_precision, copy=False
            )

        self.set_tie_point_indices(tie_point_indices, copy=copy)
        self.set_parameters(parameters, copy=copy)
        self.set_parameter_dimensions(parameter_dimensions, copy=copy)
        self.set_dependent_tie_points(dependent_tie_points, copy=copy)
        self.set_dependent_tie_point_dimensions(
            dependent_tie_point_dimensions, copy=copy
        )

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed data as an independent
        numpy array.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        # If the first or last element is requested then we don't need
        # to uncompress
        try:
            return self._first_or_last_element(indices)
        except IndexError:
            pass

        # ------------------------------------------------------------
        # Method: Uncompress the entire array and then subspace it
        # ------------------------------------------------------------
        # Initialise the un-sliced uncompressed array
        u = np.ma.masked_all(self.shape, dtype=self.dtype)

        Subarray = self.get_Subarray()
        subarray_kwargs = {
            **self.conformed_data(),
            **self.subarray_parameters(),
        }

        # Interpolate the tie points for each interpolation subarea
        for u_indices, u_shape, c_indices, subarea_indices, first, _ in zip(
            *self.subarrays()
        ):
            subarray = Subarray(
                indices=c_indices,
                shape=u_shape,
                first=first,
                subarea_indices=subarea_indices,
                **subarray_kwargs,
            )
            u[u_indices] = subarray[...]

        u = netcdf_indexer(
            u,
            mask=False,
            unpack=False,
            always_masked_array=False,
            orthogonal_indexing=True,
            copy=False,
        )
        return u[indices]

    def _conformed_dependent_tie_points(self):
        """Return the dependent tie points.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_conformed_parameters`, `conformed_data`

        :Returns:

            `dict`
                The conformed dependent tie points, keyed by their
                identities.

        """
        dependent_tie_points = self.get_dependent_tie_points()

        if dependent_tie_points:
            dependent_tie_point_dimensions = (
                self.get_dependent_tie_point_dimensions()
            )

            tp_ndim = self.source().ndim
            dims = list(range(tp_ndim))

            for identity, tp in dependent_tie_points.items():
                tp = tp.data
                tp_dims = dependent_tie_point_dimensions[identity]
                if sorted(tp_dims) == dims:
                    # The dependent tie point dimensions are already
                    # in the correct order
                    tp = tp.copy()
                else:
                    new_order = [
                        tp_dims.index(i) for i in dims if i in tp_dims
                    ]
                    tp = tp.transpose(new_order)

                dependent_tie_points[identity] = tp

        return dependent_tie_points

    def _conformed_interpolation_subarea_flags(self):
        """Return interpolation_subarea_flag interpolation parameters.

        See CF section 3.5 "Flags" and Appendix J "Coordinate
        Interpolation Methods".

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_conformed_parameters`

        :Returns:

            `dict`
                The conformed interpolation_subarea_flag parameters,
                keyed by flag names.

        """
        out = {}

        parameter = self.get_parameters().get("interpolation_subarea_flags")
        if parameter is None:
            return out

        flag_values = parameter.get_property("flag_values", None)
        flag_masks = parameter.get_property("flag_masks", None)
        flag_meanings = parameter.get_property("flag_meanings", None)

        if flag_meanings is None:
            # The interpolation_subarea_flags interpolation parameter
            # does not have the "flag_meanings" property
            return out

        if flag_masks is None and flag_values is None:
            # The interpolation_subarea_flags interpolation parameter
            # does not have the "flag_values" nor the "flag_masks"
            # properties
            return out

        flag_meanings = flag_meanings.split()

        parameter = parameter.data

        for name in self._flag_names:
            if name not in flag_meanings:
                # The interpolation_subarea_flags interpolation
                # parameter does not have this flag name in its
                # "flag_meanings" property
                continue

            flag_meanings = np.atleast_1d(flag_meanings)
            index = np.where(flag_meanings == name)[0]

            if flag_masks is not None:
                mask = np.atleast_1d(flag_masks)[index]

            if flag_values is not None:
                value = np.atleast_1d(flag_values)[index]
                if flag_masks is None:
                    # flag_values but no flag_masks
                    data = parameter == value
                else:
                    # flag_values and flag_masks
                    data = parameter & (mask & value)
            else:
                # flag_masks but no flag_values
                data = parameter & mask

            out[name] = data

        return out

    def _conformed_parameters(self):
        """Return interpolation parameters.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `_conformed_dependent_tie_points`,
                     `conformed_data`

        :Returns:

            `dict`
                The conformed parameters, keyed by their parameter
                term names.

        """
        parameters = self.get_parameters()

        if parameters:
            parameter_dimensions = self.get_parameter_dimensions()

            tp_ndim = self.source().ndim
            dims = list(range(tp_ndim))

            for term, parameter in parameters.items():
                parameter = parameter.data
                parameter_dims = parameter_dimensions[term]
                if sorted(parameter_dims) == dims:
                    # The interpolation parameter dimensions are
                    # already in the correct order
                    parameter = parameter.copy()
                else:
                    new_order = [
                        parameter_dims.index(i)
                        for i in dims
                        if i in parameter_dims
                    ]
                    parameter = parameter.transpose(new_order)

                    # Add missing interpolation parameter dimensions as
                    # size 1 axes
                    if len(parameter_dims) < tp_ndim:
                        for d in sorted(set(dims).difference(parameter_dims)):
                            parameter = parameter.insert_dimension(position=d)

                parameters[term] = parameter

            parameters.update(self._conformed_interpolation_subarea_flags())
            parameters.pop("interpolation_subarea_flags", None)

        return parameters

    @cached_property
    def bounds(self):
        """True if the compressed array represents bounds tie points.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self.ndim > self.source().ndim

    @cached_property
    def dtype(self):
        """Data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return np.dtype(float)

    def conformed_data(self):
        """Return the conformed tie points and any ancillary data.

        Returns the tie points and any ancillary data in the forms
        required by the interpolation algorithm.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `dict`
                The conformed tie points, with the key ``'data'``; the
                conformed interpolation parameters, with the key
                ``'parameters'``; and the conformed dependent tie
                points, with the key ``'dependent_tie_points'``.

        """
        out = super().conformed_data()
        out["parameters"] = self._conformed_parameters()
        out["dependent_tie_points"] = self._conformed_dependent_tie_points()
        return out

    def get_computational_precision(self, default=ValueError()):
        """Get the validation computational precision.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                computational precision has not been set.

                {{default Exception}}

        """
        out = self._get_component("computational_precision", None)
        if out is None:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no "
                "'computational_precision'",
            )

        return out

    def get_dependent_tie_point_dimensions(self):
        """Get the dimension order of dependent tie points.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_dependent_tie_points`

        :Returns:

            `dict`
                The tie point array dimensions that correspond to the
                dimensions of the dependent tie points arrays, keyed
                the dependent tie point identities.

        """
        return self._get_component("dependent_tie_point_dimensions").copy()

    def get_dependent_tie_points(self):
        """Get the dependent tie points.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_dependent_tie_point_dimensions`

        :Returns:

            `dict`
                A dictionary of dependent tie points, each keyed by
                its identity.

        """
        return self._get_component("dependent_tie_points").copy()

    def get_interpolation_description(self, default=ValueError()):
        """Get the non-standardised interpolation method description.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_interpolation_name`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                interpolation description has not been set.

                {{default Exception}}

        """
        out = self._get_component("interpolation_description", None)
        if out is None:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no "
                "'interpolation_description'",
            )

        return out

    def get_interpolation_name(self, default=ValueError()):
        """Get the name of a standardised interpolation method.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_interpolation_description`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                interpolation name has not been set.

                {{default Exception}}

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_interpolation_description`

        """
        out = self._get_component("interpolation_name", None)
        if out is None:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no 'interpolation_name'",
            )

        return out

    def get_parameter_dimensions(self):
        """Get the dimension order interpolation parameters.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_parameters`, `set_parameter_dimensions`

        :Returns:

            `dict`
                The interpolation parameter dimension orders relative
                to the tie points array, each keyed by the interpolation
                parameter term name.

        """
        return self._get_component("parameter_dimensions").copy()

    def get_parameters(self):
        """Get the interpolation parameter variables.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_parameter_dimensions`, `set_parameters`

        :Returns:

            `dict`
                The interpolation parameters, each keyed by the
                interpolation parameter term name.

        """
        return self._get_component("parameters").copy()

    def get_Subarray(self):
        """Return the Subarray class.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `Subarray`
                The class for representing subarrays.

        """
        interpolation_name = self.get_interpolation_name(None)
        Subarray = self._Subarray.get(interpolation_name)
        if Subarray is None:
            if interpolation_name is None:
                raise IndexError(
                    "Can't get Subarray class when no interpolation_name "
                    "has been set"
                )

            raise ValueError(
                "Can't get Subarray class when with unknown "
                f"interpolation_name {interpolation_name!r}"
            )

        if interpolation_name is None:
            Subarray = partial(
                Subarray,
                interpolation_description=self.get_interpolation_description(
                    ""
                ),
            )

        return Subarray

    def get_tie_point_indices(self):
        """Get the tie point index variables for subsampled dimensions.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `set_tie_point_indices`

        :Returns:

            `dict`
                The tie point index variables, each keyed by its
                corresponding subsampled dimension position. If no tie
                point index variables have been set then an empty
                dictionary is returned.

        """
        return self._get_component("tie_point_indices").copy()

    def set_dependent_tie_point_dimensions(self, value, copy=True):
        """Set the dimension order of dependent tie points.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `set_dependent_tie_points`

        :Parameters:

            value: `dict`
                The tie point array dimensions that correspond to the
                dimensions of the dependent tie points arrays, keyed
                the dependent tie point identities.

        :Returns:

            `None`

        """
        self._set_component(
            "dependent_tie_point_dimensions", value.copy(), copy=False
        )

    def set_dependent_tie_points(self, value, copy=True):
        """Set the dependent tie points.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `set_dependent_tie_point_dimensions`

        :Parameters:

            value: `dict`
                A dictionary of dependent tie points, each keyed by
                its identity.

            copy: `bool`, optional
                If False then the dependent tie points are not copies.
                By default the tie points are deep copied.

        :Returns:

            `None`

        """
        if copy:
            value = {identity: tp.copy() for identity, tp in value.items()}
        else:
            value = value.copy()

        self._set_component("dependent_tie_points", value, copy=False)

    def set_parameter_dimensions(self, value, copy=True):
        """Set the dimension order interpolation parameters.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_parameter_dimensions`

        :Parameters:

            value: `dict`
                The interpolation parameter dimension orders relative
                to the tie points array, each keyed by the interpolation
                parameter term name.

        :Returns:

            `None`

        """
        self._set_component("parameter_dimensions", value.copy(), copy=False)

    def set_parameters(self, value, copy=True):
        """Set the interpolation parameter variables.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_parameters`

        :Parameters:

            value: `dict`
                The interpolation parameters, each keyed by the
                interpolation parameter term name.

        :Returns:

            `None`

        """
        if copy:
            value = {term: p.copy() for term, p in value.items()}
        else:
            value = value.copy()

        self._set_component("parameters", value, copy=False)

    def set_tie_point_indices(self, value, copy=True):
        """Set the tie point index variables for subsampled dimensions.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_tie_point_indices`

        :Parameters:

            value: `dict`
                The tie point index variables, each keyed by its
                corresponding subsampled dimension position.

        :Returns:

            `None`

        """
        if copy:
            value = {key: i.copy() for key, i in value.items()}
        else:
            value = value.copy()

        self._set_component("tie_point_indices", value, copy=False)

    def subarray_shapes(self, shapes):
        """Create the subarray shapes along each uncompressed dimension.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `subarray`

        :Parameters:

            {{subarray_shapes chunks: `int`, sequence, `dict`, or `str`, optional}}

        :Returns:

            `list`
                The subarray sizes along each uncompressed dimension.

        >>> a.shape
        (4, 20, 30)
        >>> a.compressed_dimensions()
        {1: (1,), 2: (2,)}
        >>> a.subarray_shapes(-1)
        [(4,), None, None]
        >>> a.subarray_shapes("auto")
        ["auto", None, None]
        >>> a.subarray_shapes(2)
        [2, None, None]
        >>> a.subarray_shapes("60B")
        ["60B", None, None]
        >>> a.subarray_shapes((2, None, None))
        [2, None, None]
        >>> a.subarray_shapes(((1, 3), None, None))
        [(1, 3), None, None]
        >>> a.subarray_shapes(("auto", None, None))
        ["auto", None, None]
        >>> a.subarray_shapes(("60B", None, None))
        ["60B", None, None]

        """
        u_dims = self.get_compressed_axes()

        if shapes == -1:
            return [
                None if i in u_dims else (size,)
                for i, size in enumerate(self.shape)
            ]

        if isinstance(shapes, (str, Number)):
            return [None if i in u_dims else shapes for i in range(self.ndim)]

        if isinstance(shapes, dict):
            shapes = [
                shapes[i] if i in shapes else None for i in range(self.ndim)
            ]
        elif len(shapes) != self.ndim:
            # chunks is a sequence
            raise ValueError(
                f"Wrong number of 'shapes' elements in {shapes}: "
                f"Got {len(shapes)}, expected {self.ndim}"
            )

        # chunks is a sequence
        return [None if i in u_dims else c for i, c in enumerate(shapes)]

    def subarrays(self, shapes=-1):
        """Return descriptors for every subarray.

        These descriptors are used during subarray decompression.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            {{subarrays chunks: ``-1`` or sequence, optional}}

        :Returns:

            6-`tuple` of iterators
               Each iterator iterates over a particular descriptor
               from each subarray.

               1. The indices of the uncompressed array that
                  correspond to each interpolation subarea.

                  .. note:: If a tie point is shared with a previous
                            interpolation subarea, then that location
                            is excluded from the index, thereby
                            avoiding any overlaps between indices.

               2. The shape of each uncompressed interpolation
                  subarea, excluding the tie point locations defined
                  in previous interpolation subareas.

               3. The indices of the tie point array that correspond
                  to each interpolation subarea. Each index for the
                  tie point interpolated dimensions is expressed as a
                  list of two integers, rather than a `slice` object,
                  to facilitate retrieval of each tie point
                  individually.

               4. The index of each interpolation subarea along the
                  interpolation subarea dimensions.

               5. Flags which state, for each interpolated dimension,
                  whether each interpolation subarea is at the start of
                  a continuous area.

               6. The location of each subarray on the uncompressed
                  dimensions. Note that if the tie points are bounds
                  tie points then the uncompressed dimensions include
                  the trailing bounds dimension.

        **Examples**

        An original 3-d array with shape (20, 12, 15) has been
        compressed by subsampling with dimensions 0 and 2 being
        interpolated dimensions. Interpolated dimension 0 (size 20)
        has two equally-sized continuous areas, each with one
        interpolation subarea of size 10; and interpolated dimension 2
        (size 15) has a single continuous area divided into has three
        interpolation subareas of sizes 5, 6, and 6.

        >>> (
        ...  u_indices,
        ...  u_shapes,
        ...  c_indices,
        ...  interpolation_subarea_indices,
        ...  new_continuous_area,
        ...  subarray_locations,
        ... ) = x.subarrays()
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 10, None), slice(0, 12, None), slice(0, 5, None))
        (slice(0, 10, None), slice(0, 12, None), slice(5, 10, None))
        (slice(0, 10, None), slice(0, 12, None), slice(10, 15, None))
        (slice(10, 20, None), slice(0, 12, None), slice(0, 5, None))
        (slice(10, 20, None), slice(0, 12, None), slice(5, 10, None))
        (slice(10, 20, None), slice(0, 12, None), slice(10, 15, None))
        >>> for i in u_shapes:
        ...    print(i)
        ...
        (10, 12, 5)
        (10, 12, 5)
        (10, 12, 5)
        (10, 12, 5)
        (10, 12, 5)
        (10, 12, 5)
        >>> for i in c_indices,
        ...    print(i)
        ...
        (slice(0, 2, None), slice(0, 12, None), slice(0, 2, None))
        (slice(0, 2, None), slice(0, 12, None), slice(1, 3, None))
        (slice(0, 2, None), slice(0, 12, None), slice(2, 4, None))
        (slice(2, 4, None), slice(0, 12, None), slice(0, 2, None))
        (slice(2, 4, None), slice(0, 12, None), slice(1, 3, None))
        (slice(2, 4, None), slice(0, 12, None), slice(2, 4, None))
        >>> for i in interpolation_subarea_indices:
        ...    print(i)
        ...
        (slice(0, 1, None), slice(0, 12, None), slice(0, 1, None)
        (slice(0, 1, None), slice(0, 12, None), slice(1, 2, None)
        (slice(0, 1, None), slice(0, 12, None), slice(2, 3, None)
        (slice(1, 2, None), slice(0, 12, None), slice(0, 1, None)
        (slice(1, 2, None), slice(0, 12, None), slice(1, 2, None)
        (slice(1, 2, None), slice(0, 12, None), slice(2, 3, None)
        >>> for i in new_continuous_area:
        ...    print(i)
        ...
        (True, None, True)
        (True, None, False)
        (True, None, False)
        (True, None, True)
        (True, None, False)
        (True, None, False)
        >>> for i in subarray_locations:
        ...    print(i)
        ...
        (0, 0, 0)
        (0, 0, 1)
        (0, 0, 2)
        (1, 0, 0)
        (1, 0, 1)
        (1, 0, 2)

        """
        tie_points = self.source()
        tie_point_indices = self.get_tie_point_indices()
        u_dims = tuple(tie_point_indices)

        shapes = self.subarray_shapes(shapes)

        # The indices of the uncompressed array that correspond to
        # each interpolation subarea.
        #
        # .. note:: If a tie point is shared with a previous
        #           interpolation subarea, then that location is
        #           excluded from the index, thereby avoiding any
        #           overlaps.
        u_indices = [None] * self.ndim
        u_shapes = u_indices[:]
        subarray_locations = u_indices[:]
        for d, c in enumerate(shapes):
            if d in u_dims:
                continue

            subarray_locations[d] = [i for i in range(len(c))]
            u_shapes[d] = c

            c = tuple(accumulate((0,) + c))
            u_indices[d] = [slice(i, j) for i, j in zip(c[:-1], c[1:])]

        # The shape of each uncompressed interpolated subarea along
        # the interpolated dimensions, excluding the tie point
        # locations defined in previous interpolation subareas.
        # u_shapes = [(n,) for n in self.shape]

        # The indices of the tie point array that correspond to each
        # interpolation subarea.
        #        c_indices = [(slice(None),)] * tp_ndim
        c_indices = [(slice(0, size),) for size in tie_points.shape]

        # The location of each subarray
        # subarray_locations = [(0,)] * self.ndim

        # The index of each interpolation subarea along the
        # interpolation subarea dimensions
        interpolation_subarea_indices = c_indices[:]

        # The flags which state, for each dimension, whether (`True`)
        # or not (`False`) an interpolation subarea is at the start of
        # a continuous area. Non-interpolated dimensions are given the
        # falsey flag `None`.
        new_continuous_area = [(None,)] * tie_points.ndim

        for d, tie_point_index in tie_point_indices.items():
            u_index = []
            u_shape = []
            c_index = []
            continuous_area = []
            interpolation_subarea_index = []
            location = []

            indices = np.array(tie_point_index).tolist()

            first = True

            # Initialise the count alng the interpolation subarea
            # dimension
            j = 0
            for i, (index0, index1) in enumerate(
                zip(indices[:-1], indices[1:])
            ):
                continuous_area.append(first)

                if index1 - index0 <= 1:
                    first = True
                    continuous_area.pop()
                    continue

                # The index of the interpolation subarea along the
                # corresponding interpolated subarea dimension
                location.append(j)
                interpolation_subarea_index.append(slice(j, j + 1))
                j += 1

                # The subspace for the axis of the tie points that
                # corresponds to this axis of the interpolation
                # subarea
                c_index.append(slice(i, i + 2))

                # The size of the interpolation subarea along this
                # interpolated dimension
                size = index1 - index0 + 1
                if not first:
                    size -= 1

                u_shape.append(size)

                # The subspace for this axis of the uncompressed array
                # that corresponds to the interpolation suabrea, only
                # including the first tie point if this the first
                # interpolation subarea in a new continuous area.
                if not first:
                    index0 += 1

                u_index.append(slice(index0, index1 + 1))

                first = False

            u_indices[d] = u_index
            u_shapes[d] = u_shape
            c_indices[d] = c_index
            interpolation_subarea_indices[d] = interpolation_subarea_index
            new_continuous_area[d] = continuous_area
            subarray_locations[d] = location

        return (
            product(*u_indices),
            product(*u_shapes),
            product(*c_indices),
            product(*interpolation_subarea_indices),
            product(*new_continuous_area),
            product(*subarray_locations),
        )

    def to_dask_array(self, chunks="auto"):
        """Convert the data to a `dask` array.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            chunks: `int`, `tuple`, `dict` or `str`, optional
                Specify the chunking of the returned dask array.

                Any value accepted by the *chunks* parameter of the
                `dask.array.from_array` function is allowed.

                The chunk sizes implied by *chunks* for a dimension that
                has been fragmented are ignored and replaced with values
                that are implied by that dimensions fragment sizes.

        :Returns:

            `dask.array.Array`
                The `dask` array representation.

        """
        import dask.array as da
        from dask import config
        from dask.array.core import getter, normalize_chunks
        from dask.base import tokenize

        name = (f"{self.__class__.__name__}-{tokenize(self)}",)

        dtype = self.dtype

        context = partial(config.set, scheduler="synchronous")

        compressed_dimensions = self.compressed_dimensions()
        conformed_data = self.conformed_data()
        compressed_data = conformed_data["data"]
        parameters = conformed_data["parameters"]
        dependent_tie_points = conformed_data["dependent_tie_points"]

        # If possible, convert the compressed data, parameters and
        # dependent tie points to dask arrays that don't support
        # concurrent reads. This prevents "compute called by compute"
        # failures problems at compute time.
        #
        # TODO: This won't be necessary if this is refactored so that
        #       arrays are part of the same dask graph as the
        #       compressed subarrays.
        compressed_data = self._lock_file_read(compressed_data)
        parameters = {
            k: self._lock_file_read(v) for k, v in parameters.items()
        }
        dependent_tie_points = {
            k: self._lock_file_read(v) for k, v in dependent_tie_points.items()
        }

        # Get the (cfdm) subarray class
        Subarray = self.get_Subarray()
        subarray_name = Subarray().__class__.__name__

        # Set the chunk sizes for the dask array
        #
        # Note: The chunks created here are incorrect for the
        #       compressed dimensions, since these chunk sizes are a
        #       function of the tie point indices which haven't yet
        #       been accessed. Therefore, the chunks for the
        #       compressed dimensions need to be redefined later.
        chunks = normalize_chunks(
            self.subarray_shapes(chunks),
            shape=self.shape,
            dtype=dtype,
        )

        # Re-initialise the chunks
        u_dims = list(compressed_dimensions)
        chunks = [[] if i in u_dims else c for i, c in enumerate(chunks)]

        # For each dimension, initialise the index of the chunk
        # previously created (prior to the chunk currently being
        # created). The value -1 is an arbitrary negative value that is
        # always less than any chunk index, which is always a natural
        # number.
        previous_chunk_location = [-1] * len(chunks)

        dsk = {}
        for (
            u_indices,
            u_shape,
            c_indices,
            subarea_indices,
            first,
            chunk_location,
        ) in zip(*self.subarrays(shapes=chunks)):
            subarray = Subarray(
                data=compressed_data,
                indices=c_indices,
                shape=u_shape,
                compressed_dimensions=compressed_dimensions,
                first=first,
                subarea_indices=subarea_indices,
                parameters=parameters,
                dependent_tie_points=dependent_tie_points,
                context_manager=context,
            )

            key = f"{subarray_name}-{tokenize(subarray)}"
            dsk[key] = subarray
            dsk[name + chunk_location] = (
                getter,
                key,
                Ellipsis,
                False,
                False,
            )

            # Add correct chunk sizes for compressed dimensions
            for d in u_dims[:]:
                previous = previous_chunk_location[d]
                new = chunk_location[d]
                if new > previous:
                    chunks[d].append(u_shape[d])
                    previous_chunk_location[d] = new
                elif new < previous:
                    # No more chunk sizes required for this compressed
                    # dimension
                    u_dims.remove(d)

        chunks = [tuple(c) for c in chunks]

        # Return the dask array
        return da.Array(dsk, name[0], chunks=chunks, dtype=dtype)

    def to_memory(self):
        """Bring data on disk into memory.

        There is no change to data that is already in memory.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `{{class}}`
                A copy of the array with all of its data in memory.

        """
        a = super().to_memory()

        d = {}
        for key, v in a.get_tie_point_indices().items():
            try:
                v = v.to_memory()
            except AttributeError:
                pass

            d[key] = v

        a.set_tie_point_indices(d, copy=False)

        d = {}
        for key, v in a.get_parameters().items():
            try:
                v = v.to_memory()
            except AttributeError:
                pass

            d[key] = v

        a.set_parameters(d, copy=False)

        d = {}
        for key, v in a.get_dependent_tie_points().items():
            try:
                v = v.to_memory()
            except AttributeError:
                pass

            d[key] = v

        a.set_dependent_tie_points(d, copy=False)

        return a



import logging
from itertools import zip_longest
from re import Pattern

from . import core, mixin
from .core.functions import deepcopy
from .decorators import _manage_log_level_via_verbosity

logger = logging.getLogger(__name__)


class Constructs(mixin.Container, core.Constructs):
    """A container for metadata constructs.

    The container has similarities to a `dict` in that it presents the
    metadata constructs as key/value pairs, where the key is the
    unique identifier that corresponds to a metadata construct value;
    is indexable by metadata construct identifier; and provides a
    subset of the usual dictionary methods: `get`, `items`, `keys`,
    and `values`. The number of constructs (which may be zero) can be
    found via the Python `len` function. The container can be
    converted to an actual `dict` with the `todict` method.

    **Filtering**

    A subset of the metadata constructs can be defined and returned in
    a new `Constructs` instance by using the various filter methods.
    See `filter` for more details.

    **Calling**

    Calling a `Constructs` instance also creates a new `Constructs`
    instance that contains a subset of the metadata constructs,
    primarily selecting by construct identity. For instance, selecting
    constructs that have an identity of 'latitude' could be done by
    either ``x = c('latitude')`` or ``x =
    c.filter_by_identity('latitude')``. More generally,
    ``c(*identities, **filter_kwargs)`` is equivalent to
    ``c.filter(filter_by_identity=identities, **filter_kwargs)``.

    **Metadata constructs**

    The following metadata constructs can be included:

    * domain axis constructs
    * dimension coordinate constructs
    * auxiliary coordinate constructs
    * coordinate reference constructs
    * domain ancillary constructs
    * cell measure constructs
    * domain topology constructs
    * cell connectivity constructs
    * cell method constructs
    * field ancillary constructs

    .. versionadded:: (cfdm) 1.7.0

    """

    def __call__(self, *identities, **filter_kwargs):
        """Select metadata constructs by identity.

        Calling a `Constructs` instance selects metadata constructs by
        identity and is an alias for the `filter_by_identity`
        method. For example, to select constructs that have an
        identity of 'air_temperature': ``d = c('air_temperature')``.

        Note that ``c(*identities, **filter_kwargs)`` is equivalent to
        ``c.filter(filter_by_identity=identities, **filter_kwargs)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter_by_identity`, `filter`

        :Parameters:

            identities: optional
                Select constructs that have an identity, defined by
                their `!identities` methods, that matches any of the
                given values.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs`
                The selected constructs and their construct keys.

        **Examples**

        See `filter_by_identity` and `filter` for examples.

        """
        if identities:
            if "filter_by_identity" in filter_kwargs:
                raise TypeError(
                    f"Can't set {self.__class__.__name__}.__call__() "
                    "keyword argument 'filter_by_identity' when "
                    "positional *identities arguments are also set"
                )

            # Ensure that filter_by_identity is the last filter
            # applied, as it's the most expensive.
            filter_kwargs["filter_by_identity"] = identities

        return self.filter(**filter_kwargs)

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        ignore = self._ignore
        construct_types = [
            f"{c}({len(v)})"
            for c, v in sorted(self._constructs.items())
            if len(v) and c not in ignore
        ]

        return f"<{self.__class__.__name__}: {', '.join(construct_types)}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        .. versionadded:: (cfdm) 1.7.0

        """
        out = ["Constructs:"]

        ignore = self._ignore
        construct_types = [
            c
            for c, v in sorted(self._constructs.items())
            if len(v) and c not in ignore
        ]

        first = "{"
        for construct_type in construct_types:
            for key, value in sorted(self._constructs[construct_type].items()):
                if first:
                    out[0] = out[0] + f"\n{{{key!r}: {value!r},"
                    first = False
                else:
                    out.append(f"{key!r}: {value!r},")

        if first:
            out[0] = out[0] + "\n{}"
        else:
            out[-1] = out[-1][:-1] + "}"

        return "\n ".join(out)

    def _axes_to_constructs(self):
        """Maps domain axes to constructs whose data span them.

        This is useful for ascertaining whether or not two `Constructs`
        instances are equal.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `dict`

        **Examples**

        >>> f.constructs._axes_to_constructs()
        {('domainaxis0',): {'auxiliary_coordinate': {},
                            'cell_measure'        : {},
                            'dimension_coordinate': {'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >},
                            'domain_ancillary'    : {'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
                                                     'domainancillary1': <DomainAncillary: ncvar%b(1) >},
                            'domain_topology'     : {},
                            'cell_connectivity'   : {},
                            'field_ancillary'     : {}},
         ('domainaxis1',): {'auxiliary_coordinate': {'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name=Grid latitude name(10) >},
                            'cell_measure'        : {},
                            'dimension_coordinate': {'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>},
                            'domain_ancillary'    : {},
                            'domain_topology'     : {},
                            'cell_connectivity'   : {},
                            'field_ancillary'     : {}},
         ('domainaxis1', 'domainaxis2'): {'auxiliary_coordinate': {'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>},
                                          'cell_measure'        : {},
                                          'dimension_coordinate': {},
                                          'domain_ancillary'    : {'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>},
                                          'domain_topology'     : {},
                                          'cell_connectivity'   : {},
                                          'field_ancillary'     : {'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}},
         ('domainaxis2',): {'auxiliary_coordinate': {},
                            'cell_measure'        : {},
                            'dimension_coordinate': {'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>},
                            'domain_ancillary'    : {},
                            'domain_topology'     : {},
                            'cell_connectivity'   : {},
                            'field_ancillary'     : {}},
         ('domainaxis2', 'domainaxis1'): {'auxiliary_coordinate': {'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>},
                                          'cell_measure'        : {'cellmeasure0': <CellMeasure: measure:area(9, 10) km2>},
                                          'dimension_coordinate': {},
                                          'domain_ancillary'    : {},
                                          'domain_topology'     : {},
                                          'cell_connectivity'   : {},
                                          'field_ancillary'     : {}},
         ('domainaxis3',): {'auxiliary_coordinate': {},
                            'cell_measure'        : {},
                            'dimension_coordinate': {'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >},
                            'domain_ancillary'    : {},
                            'domain_topology'     : {},
                            'cell_connectivity'   : {},
                            'field_ancillary'     : {}}}

        """
        data_axes = self.data_axes()
        array_constructs = self._array_constructs
        construct_type = self._construct_type
        constructs = self._constructs

        ignore = self._ignore

        out = {
            axes: {
                ctype: {}
                for ctype in array_constructs
                if ctype not in ignore and constructs.get(ctype)
            }
            for axes in data_axes.values()
        }

        for cid, construct in self.filter_by_data(todict=True).items():
            axes = data_axes.get(cid)
            out[axes][construct_type[cid]][cid] = construct

        return out

    def _del_construct(self, key, default=ValueError()):
        """Remove a metadata construct.

        If a domain axis construct is selected for removal then it can't
        be spanned by any metdata construct data arrays, nor be referenced
        by any cell method constructs.

        However, a domain ancillary construct may be removed even if it is
        referenced by coordinate reference construct. In this case the
        reference is replace with `None`.

        If a climatological time cell method construct is removed then the
        climatological status of its corresponding coordinate constructs
        will be reviewed.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_get_construct`, `_set_construct`

        :Parameters:

            key: `str`
                The key of the construct to be removed.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                construct can not be removed, or does not exist.

                {{default Exception}}

        :Returns:

                The removed construct.

        **Examples**

        >>> x = c._del_construct('auxiliarycoordinate2')

        """
        out = super()._del_construct(key, default=default)

        try:
            is_cell_method = out.construct_type == "cell_method"
        except AttributeError:
            is_cell_method = False

        if is_cell_method:
            # --------------------------------------------------------
            # Since a cell method construct was deleted, check to see
            # if it was for climatological time, and if so reset the
            # climatology status of approriate coordinate constructs.
            # --------------------------------------------------------
            qualifiers = out.qualifiers()
            if "within" in qualifiers or "over" in qualifiers:
                axes = out.get_axes(default=())
                if len(axes) == 1 and axes[0] not in self.filter_by_type(
                    "domain_axis", todict=True
                ):
                    coordinates = {}
                    for ckey, c in self.filter_by_type(
                        "dimension_coordinate",
                        "auxiliary_coordinate",
                        todict=True,
                    ).items():
                        if axes != self.data_axes().get(ckey, ()):
                            continue

                        # This coordinate construct spans the deleted
                        # cell method axes, so unset its climatology
                        # setting.
                        c.set_climatology(False)
                        coordinates[ckey] = c

                    # Reset the climatology settings of all the
                    # modified coordinate constructs. Do this because
                    # there may still be non-deleted cell methods for
                    # the same axes that still define climatological
                    # time.
                    if coordinates:
                        self._set_climatology(coordinates=coordinates)

        return out

    @_manage_log_level_via_verbosity
    def _equals_cell_method(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_type=False,
        axis1_to_axis0=None,
        key1_to_key0=None,
    ):
        """Whether two cell method constructs are the same."""
        cell_methods0 = self._construct_dict("cell_method")
        cell_methods1 = other._construct_dict("cell_method")

        len0 = len(cell_methods0)
        if len0 != len(cell_methods1):
            logger(
                "Different numbers of cell methods: "
                f"{cell_methods0!r} != {cell_methods1!r}"
            )  # pragma: no cover
            return False

        if not len0:
            return True

        axis0_to_axis1 = {
            axis0: axis1 for axis1, axis0 in axis1_to_axis0.items()
        }

        for cm0, cm1 in zip(
            tuple(cell_methods0.values()), tuple(cell_methods1.values())
        ):
            # Check that there are the same number of axes
            axes0 = cm0.get_axes(())
            axes1 = list(cm1.get_axes(()))
            if len(axes0) != len(axes1):
                logger.info(
                    f"{cm0.__class__.__name__}: Different cell methods "
                    f"(mismatched axes):\n  {cell_methods0}\n  {cell_methods1}"
                )  # pragma: no cover
                return False

            indices = []
            for axis0 in axes0:
                if axis0 is None:
                    return False
                for axis1 in axes1:
                    if axis0 in axis0_to_axis1 and axis1 in axis1_to_axis0:
                        if axis1 == axis0_to_axis1[axis0]:
                            # axis0 and axis1 are domain axis
                            # constructs that correspond to each other
                            axes1.remove(axis1)
                            indices.append(cm1.get_axes(()).index(axis1))
                            break
                    elif axis0 in axis0_to_axis1 or axis1 in axis1_to_axis0:
                        # Only one of axis0 and axis1 is a domain axis
                        # construct
                        logger.info(
                            f"{cm0.__class__.__name__}: Different cell "
                            "methods (mismatched axes):\n  "
                            f"{cell_methods0}\n  {cell_methods1}"
                        )  # pragma: no cover
                        return False
                    elif axis0 == axis1:
                        # axes0 and axis 1 are identical standard
                        # names
                        axes1.remove(axis1)
                        indices.append(cm1.get_axes(()).index(axis1))
                    elif axis1 is None:
                        # axis1
                        logger.info(
                            f"{cm0.__class__.__name__}: Different cell "
                            "methods (mismatched axes):\n  "
                            f"{cell_methods0}\n  {cell_methods1}"
                        )  # pragma: no cover
                        return False

            if len(cm1.get_axes(())) != len(indices):
                logger.info(
                    f"{cm0.__class__.__name__}: [4] Different cell methods "
                    f"(mismatched axes):\n  {cell_methods0}\n  {cell_methods1}"
                )  # pragma: no cover
                return False

            cm1 = cm1.sorted(indices=indices)
            cm1.set_axes(axes0)

            if not cm0.equals(
                cm1,
                atol=atol,
                rtol=rtol,
                verbose=verbose,
                ignore_type=ignore_type,
            ):
                logger.info(
                    f"{cm0.__class__.__name__}: Different cell methods: "
                    f"{cell_methods0!r}, {cell_methods1!r}"
                )  # pragma: no cover
                return False

        return True

    @_manage_log_level_via_verbosity
    def _equals_coordinate_reference(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_type=False,
        axis1_to_axis0=None,
        key1_to_key0=None,
    ):
        """Whether two coordinate reference constructs are the same."""
        refs0 = self._construct_dict("coordinate_reference")
        refs1 = other._construct_dict("coordinate_reference", copy=True)

        if len(refs0) != len(refs1):
            logger.info(
                f"{self.__class__.__name__}: Different numbers of coordinate "
                f"reference constructs: {len(refs0)} != {len(refs1)}"
            )  # pragma: no cover

            return False

        if refs0:
            if verbose == -1:
                debug_verbose = 2
            else:
                debug_verbose = 0

            for ref0 in refs0.values():
                found_match = False
                for key1, ref1 in tuple(refs1.items()):
                    logger.debug(
                        f"{self.__class__.__name__}: Comparing {ref0!r}, "
                        f"{ref1!r}: "
                    )  # pragma: no cover

                    if not ref0.equals(
                        ref1,
                        rtol=rtol,
                        atol=atol,
                        verbose=debug_verbose,
                        ignore_type=ignore_type,
                    ):
                        continue

                    # Coordinates
                    coordinates0 = ref0.coordinates()
                    coordinates1 = set()
                    for value in ref1.coordinates():
                        coordinates1.add(key1_to_key0.get(value, value))

                    if coordinates0 != coordinates1:
                        logger.debug(
                            "Coordinates don't match"
                        )  # pragma: no cover

                        continue

                    # Domain ancillary-valued coordinate conversion terms
                    terms0 = ref0.coordinate_conversion.domain_ancillaries()

                    terms1 = {
                        term: key1_to_key0.get(key, key)
                        for term, key in (
                            ref1.coordinate_conversion.domain_ancillaries().items()
                        )
                    }

                    if terms0 != terms1:
                        logger.debug(
                            "Coordinate conversion domain ancillaries "
                            "don't match"
                        )  # pragma: no cover

                        continue

                    logger.debug("OK")  # pragma: no cover

                    found_match = True
                    del refs1[key1]
                    break

                if not found_match:
                    logger.info(
                        f"{self.__class__.__name__}: No match for {ref0!r}"
                    )  # pragma: no cover
                    return False

        return True

    @_manage_log_level_via_verbosity
    def _equals_domain_axis(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_type=False,
        axis1_to_axis0=None,
        key1_to_key0=None,
    ):
        """Whether two domain axes constructs are the same."""
        domain_axes = self._construct_dict("domain_axis")
        self_sizes = sorted([d.get_size() for d in domain_axes.values()])

        domain_axes = other._construct_dict("domain_axis")
        other_sizes = sorted([d.get_size() for d in domain_axes.values()])

        if self_sizes != other_sizes:
            # There is not a 1-1 correspondence between axis sizes
            logger.info(
                f"{self.__class__.__name__}: Different domain axis sizes: "
                f"{self_sizes} != {other_sizes}"
            )  # pragma: no cover
            return False

        return True

    def _set_climatology(self, cell_methods=None, coordinates=None):
        """Set the climatology flag on approriate coordinate constructs.

        The setting is based on the cell method constructs.

        .. versionadded:: (cfdm) 1.9.0.0

        :Parameters:

            cell_methods: `dict`, optional
                TODO

            coordinates: `dict`, optional
                TODO

        :Returns:

            `list` of `str`
                The domain axis construct identifiers of all
                climatological time axes. The list may contain axis
                duplications.

        """
        out = []

        domain_axes = self.filter_by_type("domain_axis", todict=True)

        if coordinates:
            cell_methods = self.filter_by_type("cell_method", todict=True)
        elif cell_methods:
            coordinates = self.filter_by_type(
                "dimension_coordinate", "auxiliary_coordinate", todict=True
            )
        else:
            cell_methods = self.filter_by_type("cell_method", todict=True)
            coordinates = self.filter_by_type(
                "dimension_coordinate", "auxiliary_coordinate", todict=True
            )

        for key, cm in cell_methods.items():
            qualifiers = cm.qualifiers()
            if not ("within" in qualifiers or "over" in qualifiers):
                continue

            axes = cm.get_axes(default=())
            if len(axes) != 1:
                continue

            axis = axes[0]
            if axis not in domain_axes:
                continue

            # Still here? Then this axis is a climatological time axis
            out.append(axis)

            # Flag the appropriate coordinates as being for
            # climatological time.
            axes = set(axes)
            for ckey, c in coordinates.items():
                if axes != set(self.data_axes().get(ckey, ())):
                    # Construct does not span the cell method axes
                    continue

                # Construct spans the correct axes and has reference
                # time units
                c.set_climatology(True)

        return out

    def _set_construct(self, construct, key=None, axes=None, copy=True):
        """Set a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_del_construct`, `_get_construct`,
                     `_set_construct_data_axes`

        :Parameters:

            construct:
                The metadata construct to be inserted.

            key: `str`, optional
                The construct identifier to be used for the construct. If
                not set then a new, unique identifier is created
                automatically. If the identifier already exists then the
                exisiting construct will be replaced.

                *Parameter example:*
                  ``key='cellmeasure0'``

            axes: (sequence of) `str`, optional
                The construct identifiers of the domain axis constructs
                spanned by the data array. An exception is raised if used
                for a metadata construct that can not have a data array,
                i.e. domain axis, cell method and coordinate reference
                constructs.

                The axes may also be set afterwards with the
                `_set_construct_data_axes` method.

                *Parameter example:*
                  ``axes='domainaxis1'``

                *Parameter example:*
                  ``axes=['domainaxis1']``

                *Parameter example:*
                  ``axes=('domainaxis1', 'domainaxis0')``

            copy: `bool`, optional
                If True then return a copy of the unique selected
                construct. By default the construct is copied.

        :Returns:

             `str`
                The construct identifier for the construct.

        **Examples**

        >>> key = f._set_construct(c)
        >>> key = f._set_construct(c, copy=False)
        >>> key = f._set_construct(c, axes='domainaxis2')
        >>> key = f._set_construct(c, key='cellmeasure0')

        """
        if copy:
            # Create a deep copy of the construct
            construct = construct.copy()

        key = super()._set_construct(construct, key=key, axes=axes, copy=False)

        # Set any appropriate climatology flags
        construct_type = construct.construct_type
        if construct_type in ("dimension_coordinate", "auxiliary_coordinate"):
            self._set_climatology(coordinates={key: construct})
        elif construct_type == "cell_method":
            self._set_climatology(cell_methods={key: construct})

        # Return the identifier of the construct
        return key

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def _filter_convert_to_domain_axis(
        self, values, check_axis_identities=True
    ):
        """Convert values to domain axis construct identifiers.

        If possible, convert each value from a field data array index
        position to domain axis construct identifiers; or 1-d
        coordinate construct identifier to a domain axis identifier of
        the spanned domain axis construct.

        .. note:: If *values* is an empty sequence then calling this
                  method has no effect, but still takes some time to
                  execute. In other words, if performance is an issue
                  then avoid calling this method when *values* is
                  empty.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            values: sequence

            check_axis_identities: `bool` optional
                If True then check for domain axis identities.

        :Returns:

            `list`
                The parsed values

        """
        # Get the original unfiltered constructs. This gives access to
        # the full collection of domain axis, dimension coordinate and
        # auxiliary coordinate constructs.
        unfiltered = self.unfilter(copy=False)
        unfiltered_domain_axes = unfiltered._constructs.get("domain_axis", ())
        unfiltered_data_axes = unfiltered._construct_axes

        field_data = self._field_data_axes

        values2 = []
        for value in values:
            if value in unfiltered_domain_axes:
                # This value is already a domain axis construct
                # identifier
                values2.append(value)
                continue

            if field_data:
                # Try to convert a field data array index position to
                # a domain axis identifier
                try:
                    value = field_data[value]
                except TypeError:
                    pass
                except IndexError:
                    continue
                else:
                    values2.append(value)
                    continue

            # Try to convert a 1-d coordinate into a domain axis
            # identifier
            c = unfiltered.filter(
                filter_by_type=(
                    "dimension_coordinate",
                    "auxiliary_coordinate",
                ),
                filter_by_naxes=(1,),
                filter_by_identity=(value,),
                todict=True,
            )
            if c:
                c_axes = [unfiltered_data_axes[key][0] for key in c]
                if len(set(c_axes)) == 1:
                    value = c_axes[0]
                    values2.append(value)

                continue

            if check_axis_identities:
                # Try to convert a domain axis identity into a domain
                # axis identifier
                c = self.filter(
                    filter_by_type=("domain_axis",),
                    filter_by_identity=(value,),
                    todict=True,
                )
                if len(c) == 1:
                    key, _ = c.popitem()
                    values2.append(key)

        return values2

    @classmethod
    def _filter_parse_args(
        cls,
        method,
        args,
        todict=False,
        axis_mode=None,
        property_mode=None,
        _identity_config=None,
    ):
        """Parse arguments destined for a filter method.

        Specifically transforms the arguments to a filter method
        (e.g. `filter_by_property`) so that they can be used in the
        corresponding underscore filter method
        (e.g. `_filter_by_property`).

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            method: `str`
                The name of a filter method (e.g. "filter_by_data").

            args: sequence or `dict`
                The input arguments for the filter method defined by
                *method*.

            todict: `bool`
                The value of the filter method's *todict* parameter.

            axis_mode: `str`, optional
                Provide a value for the *axis_mode* parameter of the
                `filter_by_axis` method.

            property_mode: `str`, optional
                Provide a value for the *property_mode* parameter of
                the `filter_by_property` method.

            _identity_config: optional
                Provide a value for the *_config* parameter of the
                `filter_by_identity` method.

        :Returns:

            `tuple`
                The parsed positional arguments for corresponding
                underscore filter method.

        """
        out = (args, todict)

        if method == "filter_by_type":
            return out

        if method == "filter_by_identity":
            out += (_identity_config,)
        elif method == "filter_by_axis":
            out += (axis_mode,)
        elif method == "filter_by_property":
            out += ((property_mode,),)

        return out

    @classmethod
    def _filter_preprocess(cls, arg, todict=False, filter_applied=None):
        """Preprocess a `dict` or `Constructs` prior to filtering.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            arg: `Constructs` or `dict`
                Either a `Constructs` object or its dictionary
                representation.

            todict: `bool`, optional
                If True and *arg* is a `Constructs` object, then
                return its dictionary representation.

            filter_applied: `dict` optional
                If *todict* is False and *arg* is a `Constructs`
                object, then add this record to the `_filters_applied`
                attribute.

        :Returns:

            `Constructs` or `dict`, and its pop method
                If *arg* is a `Constructs` object, then either a
                shallow copy or its dictionary representation is
                returned. If *arg* is a `dict` then it is returned
                unchanged. In addtion, the "pop" method of the object
                also returned.

        """
        if isinstance(arg, dict):
            return arg, arg.pop

        if todict:
            arg = arg.todict()
            return arg, arg.pop

        arg = arg.shallow_copy()

        if filter_applied:
            arg._prefiltered = arg.shallow_copy()
            arg._filters_applied = arg.filters_applied() + (filter_applied,)

        return arg, arg._pop

    @classmethod
    def _matching_values(cls, value0, construct, value1, basic=False):
        """Whether or not two values are the same.

        :Parameters:

            value0:
                The first value to be matched. Could be a `re.Pattern`
                object.

            construct:
                The construct whose `_equals` method might be used to
                determine whether values can be considered to match.

            value1:
                The second value to be matched.

            basic: `bool`
                If True then replace the expensive `_equals` method
                with ``==``. In this case, *construct* is not used and
                may take any value.

        :Returns:

            `bool`
                Whether or not the two values match.

        """
        if isinstance(value0, Pattern):
            try:
                return value0.search(value1)
            except TypeError:
                return False

        if basic:
            return value0 == value1

        return construct._equals(value1, value0)

    @classmethod
    def _short_iteration(cls, x):
        """The default short cicuit test.

        If this method returns True then only the first identity
        returned by the construct's `!identities` method will be
        checked.

        See `_filter_by_identity` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        :Parameters:

            x: `str`
                The value against which the construct's identities are
                being compared.

        :Returns:

            `bool`
                 Returns `True` if a construct's `identities` method
                 is to short circuit after the first identity is
                 computed, otherwise `False`.

        """
        return (
            isinstance(x, str)
            and "=" not in x
            and ":" not in x
            and "%" not in x
        )

    def copy(self, data=True):
        """Return a deep copy.

        ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            data: `bool`, optional
                If True (the default) then copy data contained in the
                metadata constructs.

        :Returns:

            `Constructs`
                The deep copy.

        **Examples**

        >>> g = f.copy()
        >>> g = f.copy(data=False)

        """
        out = super().copy(data=data)

        prefiltered = getattr(self, "_prefiltered", None)
        if prefiltered is not None:
            out._prefiltered = prefiltered.copy(data=data)
            out._filters_applied = self._filters_applied

        return out

    def domain_axis_identity(self, key):
        """Return the canonical identity for a domain axis construct.

        The identity is the first found of the following:

        1. The canonical identity of a dimension coordinate construct that
           span the domain axis construct.
        2. The identity of a one-dimensional auxiliary coordinate
           construct that spans the domain axis construct. This will
           either be the value of a ``standard_name``, ``cf_role``
           (preceded by ``'cf_role='``) or ``axis`` (preceded by
           ``'axis='``) property, as appropriate.
        3. The netCDF dimension name, preceded by 'ncdim%'.
        4. The domain axis construct key, preceded by 'key%'.

        :Parameters:

            key: `str`
                The construct key for the domain axis construct.

                *Parameter example:*
                  ``key='domainaxis2'``

        :Returns:

            `str`
                The identity.

        **Examples**

        >>> c.domain_axis_identity('domainaxis1')
        'longitude'
        >>> c.domain_axis_identity('domainaxis2')
        'axis=Y'
        >>> c.domain_axis_identity('domainaxis3')
        'cf_role=timeseries_id'
        >>> c.domain_axis_identity('domainaxis4')
        'key%domainaxis4')

        """
        domain_axes = self._construct_dict("domain_axis")

        if key not in domain_axes:
            raise ValueError(f"No domain axis construct with key {key!r}")

        constructs_data_axes = self.data_axes()

        # Try to get the identity from an dimension coordinate
        dimension_coordinates = self._construct_dict("dimension_coordinate")
        identity = ""
        for dkey, dim in dimension_coordinates.items():
            if constructs_data_axes.get(dkey) == (key,):
                identity = dim.identity()
                if identity.startswith("ncvar%"):
                    identity = ""

                break

        if identity:
            return identity

        # Try to get the identity from an auxiliary coordinate
        auxiliary_coordinates = self._construct_dict("auxiliary_coordinate")

        identities = []
        for akey, aux in auxiliary_coordinates.items():
            if constructs_data_axes.get(akey) == (key,):
                identity = aux.identity()
                if not identity.startswith("ncvar%"):
                    identities.append(identity)

        if len(identities) == 1:
            return identities[0]
        elif len(identities) > 1:
            cf_role = []
            axis = []
            for i in identities:
                if i.startswith("axis="):
                    axis.append(i)
                elif i.startswith("cf_role="):
                    cf_role.append(i)

            if len(cf_role) == 1:
                return cf_role[0]

            if len(axis) == 1:
                return axis[0]

        # Try to get the identity from an netCDF dimension name
        ncdim = domain_axes[key].nc_get_dimension(None)
        if ncdim is not None:
            return f"ncdim%{ncdim}"

        # Get the identity from the domain axis construct key
        return f"key%{key}"

    def domain_axes(self, *identities, **filter_kwargs):
        """Return domain axis constructs.

        .. versionadded:: (cfdm) 1.8.9.0

        .. seealso:: `constructs`

        :Parameters:

            identities: `tuple`, optional
                Select domain axis constructs that have an identity,
                defined by their `!identities` methods, that matches
                any of the given values.

                Additionally, the values are matched against construct
                identifiers, with or without the ``'key%'`` prefix.

                Additionally, if for a given ``value``,
                ``c.filter(filter_by_type=["dimension_coordinate",
                "auxiliary_coordinate"], filter_by_naxes=(1,),
                filter_by_identity=(value,))`` returns 1-d coordinate
                constructs that all span the same domain axis
                construct then that domain axis construct is
                selected. See `filter` for details.

                Additionally, if there is an associated `Field` data
                array and a value matches the integer position of an
                array dimension, then the corresponding domain axis
                construct is selected.

                If no values are provided then all domain axis
                constructs are selected.

                {{value match}}

                {{displayed identity}}

            {{filter_kwargs: optional}} Also to configure the returned value.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

                {{Returns constructs}}

        **Examples**

        """
        cached = filter_kwargs.get("cached")
        if cached is not None:
            return cached

        if filter_kwargs:
            if "filter_by_type" in filter_kwargs:
                raise TypeError(
                    "domain_axes() got an unexpected keyword argument "
                    "'filter_by_type'"
                )

            if identities:
                if "filter_by_identity" in filter_kwargs:
                    raise TypeError(
                        "Can't set domain_axes() keyword argument "
                        "'filter_by_identity' when positional *identities "
                        "arguments are also set"
                    )
            elif "filter_by_identity" in filter_kwargs:
                identities = filter_kwargs["filter_by_identity"]

        if identities:
            # Make sure that filter_by_identity is the last filter
            # applied
            filter_kwargs["filter_by_identity"] = identities

            out, keys, hits, misses = self.filter(
                filter_by_type=("domain_axis",),
                _identity_config={"return_matched": True},
                **filter_kwargs,
            )
            if out is not None:
                return out

            keys.update(
                self._filter_convert_to_domain_axis(
                    misses, check_axis_identities=False
                )
            )

            if not keys:
                # No keys were found but some criteria were provided,
                # so force filter_by_key to return no domain axis
                # constructs.
                keys = (None,)

            filter_kwargs = {
                "filter_by_key": keys,
                "todict": filter_kwargs.get("todict", False),
            }

        return self.filter(filter_by_type=("domain_axis",), **filter_kwargs)

    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_data_type=False,
        ignore_fill_value=False,
        ignore_compression=True,
        _ignore_type=False,
        _return_axis_map=False,
    ):
        """Whether two `Constructs` instances are the same.

        {{equals tolerance}}

        {{equals compression}}

        Any type of object may be tested but equality is only possible
        with another `Constructs` construct, or a subclass of one.

        {{equals netCDF}}

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            ignore_fill_value: `bool`, optional
                If True then the ``_FillValue`` and ``missing_value``
                properties are omitted from the comparison for the
                metadata constructs.

            {{ignore_data_type: `bool`, optional}}

            {{ignore_compression: `bool`, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

        :Returns:

            `bool`
                Whether the two instances are equal.

        **Examples**

        >>> x.equals(x)
        True
        >>> x.equals(x.copy())
        True
        >>> x.equals('something else')
        False

        """
        if self is other:
            if not _return_axis_map:
                return True

        # Check that each instance is the same type
        if not isinstance(other, self.__class__):
            logger.info(
                f"{self.__class__.__name__}: Incompatible type: "
                f"{other.__class__.__name__}"
            )  # pragma: no cover
            if not _return_axis_map:
                return False

        if verbose == -1:
            debug_verbose = 2
        else:
            debug_verbose = 0

        axes0_to_axes1 = {}
        axis0_to_axis1 = {}
        axis1_to_axis0 = {}
        key1_to_key0 = {}

        # ------------------------------------------------------------
        # Domain axis constructs
        # ------------------------------------------------------------
        if not self._equals_domain_axis(
            other,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_type=_ignore_type,
            axis1_to_axis0=axis1_to_axis0,
            key1_to_key0=key1_to_key0,
        ):
            if not _return_axis_map:
                return False

        # ------------------------------------------------------------
        # Constructs with arrays
        # ------------------------------------------------------------
        log = []
        axes_to_constructs0 = self._axes_to_constructs()
        axes_to_constructs1 = other._axes_to_constructs()

        if len(axes_to_constructs0) != len(axes_to_constructs1):
            logger.info(
                f"{self.__class__.__name__}: Can't match constructs"
            )  # pragma: no cover
            return False

        for axes0, constructs0 in axes_to_constructs0.items():
            matched_all_constructs_with_these_axes = False

            len_axes0 = len(axes0)
            for axes1, constructs1 in tuple(axes_to_constructs1.items()):
                constructs1 = constructs1.copy()

                if len_axes0 != len(axes1):
                    # axes1 and axes0 contain different number of
                    # domain axes.
                    continue

                for construct_type in self._array_constructs:
                    # role_constructs0 = constructs0[construct_type]
                    # role_constructs1 = constructs1[construct_type].copy()
                    role_constructs0 = constructs0.get(construct_type, {})
                    role_constructs1 = constructs1.get(
                        construct_type, {}
                    ).copy()

                    if len(role_constructs0) != len(role_constructs1):
                        # There are the different numbers of
                        # constructs of this type
                        matched_all_constructs_with_these_axes = False
                        log.append(
                            f"{self.__class__.__name__}: Different numbers "
                            f"of {construct_type} constructs: "
                            f"{len(role_constructs0)} != "
                            f"{len(role_constructs1)}"
                        )
                        break

                    # Note: the following set of log calls are not warnings
                    # as such, but set them as warnings so they only emerge
                    # at higher verbosity level than 'INFO'.

                    # Check that there are matching pairs of equal
                    # constructs
                    matched_construct = True

                    for key0, item0 in role_constructs0.items():
                        matched_construct = False
                        for key1, item1 in tuple(role_constructs1.items()):
                            logger.debug(
                                f"{self.__class__.__name__}: Comparing "
                                f"{item0!r}, {item1!r}: "
                            )  # pragma: no cover

                            if item0.equals(
                                item1,
                                rtol=rtol,
                                atol=atol,
                                verbose=debug_verbose,
                                ignore_data_type=ignore_data_type,
                                ignore_fill_value=ignore_fill_value,
                                ignore_compression=ignore_compression,
                                ignore_type=_ignore_type,
                            ):
                                logger.debug("OK")  # pragma: no cover

                                del role_constructs1[key1]
                                key1_to_key0[key1] = key0
                                matched_construct = True
                                break

                        if not matched_construct:
                            log.append(
                                f"{self.__class__.__name__}: Can't match "
                                f"{item0!r}"
                            )
                            break

                    if role_constructs1:
                        # At least one construct in other is not equal
                        # to a construct in self
                        break

                    # Still here? Then all constructs of this type
                    # that spanning these axes match
                    constructs1.pop(construct_type, None)

                matched_all_constructs_with_these_axes = not constructs1
                if matched_all_constructs_with_these_axes:
                    del axes_to_constructs1[axes1]
                    break

            if not matched_all_constructs_with_these_axes:
                names = [self.domain_axis_identity(axis0) for axis0 in axes0]
                logger.info(
                    f"{self.__class__.__name__}: Can't match constructs "
                    f"spanning axes {names}"
                )
                if log:
                    logger.info("\n".join(log))
                if not _return_axis_map:
                    return False
            else:
                # Map item axes in the two instances
                axes0_to_axes1[axes0] = axes1

        for axes0, axes1 in axes0_to_axes1.items():
            for axis0, axis1 in zip(axes0, axes1):
                if axis0 in axis0_to_axis1 and axis1 != axis0_to_axis1[axis0]:
                    logger.info(
                        f"{self.__class__.__name__}: Ambiguous axis mapping "
                        f"({self.domain_axis_identity(axes0)} -> both "
                        f"{other.domain_axis_identity(axis1)} and "
                        f"{other.domain_axis_identity(axis0_to_axis1[axis0])})"
                    )  # pragma: no cover
                    if not _return_axis_map:
                        return False
                elif (
                    axis1 in axis1_to_axis0 and axis0 != axis1_to_axis0[axis1]
                ):
                    logger.info(
                        f"{self.__class__.__name__}: Ambiguous axis mapping "
                        f"({self.domain_axis_identity(axis0)} -> both "
                        f"{self.domain_axis_identity(axis1_to_axis0[axis0])} "
                        f"and {other.domain_axis_identity(axes1)})"
                    )  # pragma: no cover
                    if not _return_axis_map:
                        return False

                axis0_to_axis1[axis0] = axis1
                axis1_to_axis0[axis1] = axis0

        if _return_axis_map:
            return axis0_to_axis1

        # ------------------------------------------------------------
        # Constructs with no arrays
        # ------------------------------------------------------------
        for construct_type in self._non_array_constructs:
            if not getattr(self, "_equals_" + construct_type)(
                other,
                rtol=rtol,
                atol=atol,
                verbose=verbose,
                ignore_type=_ignore_type,
                axis1_to_axis0=axis1_to_axis0,
                key1_to_key0=key1_to_key0,
            ):
                return False

        # ------------------------------------------------------------
        # Still here? Then the two objects are equal
        # ------------------------------------------------------------
        return True

    def filter(
        self,
        axis_mode="and",
        property_mode="and",
        todict=False,
        cached=None,
        _identity_config={},
        **filters,
    ):
        """Select metadata constructs by a chain of filters.

        This method allows multiple filters defined by the
        "filter_by_*" methods to be chained in an alternative manner
        to calling the individual methods in sequence.

        For instance, to select the domain axis constructs with size
        73 or 96

           >>> c2 = c.filter(filter_by_type=['domain_axis'],
           ...               filter_by_size=[73, 96])

        is equivalent to

           >>> c2 = c.filter_by_type('domain_axis')
           >>> c2 = c2.filter_by_size(73, 96)

        When the results are requested as a dictionary as opposed to a
        `Constructs` object (see the *todict* parameter), using the
        `filter` method to call two or more filters is faster than
        calling the individual methods in sequence. For instance

           >>> d = c.filter(filter_by_type=['dimension_coordinate'],
           ...              filter_by_identity=['time'],
           ...              todict=True)

        is equivalent to, but faster than

           >>> c2 = c.filter_by_type('dimension_coordinate')
           >>> d = c2.filter_by_identity('time', todict=True)

        .. versionadded:: (cfdm) 1.8.9.0

        .. seealso:: `filter_by_axis`, `filter_by_cell`,
                     `filter_by_connectivity`, `filter_by_data`,
                     `filter_by_identity`, `filter_by_key`,
                     `filter_by_measure`, `filter_by_method`,
                     `filter_by_identity`, `filter_by_naxes`,
                     `filter_by_ncdim`, `filter_by_ncvar`,
                     `filter_by_property`, `filter_by_type`,
                     `filters_applied`, `inverse_filter`, `unfilter`,
                     `clear_filters_applied`

        :Parameters:

            filters: optional
                Keyword arguments defining the filters to apply. Each
                filter keyword defines a filter method, and its value
                provides the arguments for that method.

                For instance, ``filter_by_type=['domain_axis']`` will
                cause the `filter_by_type` method to be called with
                positional arguments ``*['domain_axis']``.

                The filters are applied in the same order that the
                keyword arguments appear.

                Valid keywords and their values are:

                ==========================  ==========================
                Keyword                     Value
                ==========================  ==========================
                ``filter_by_axis``          A sequence as expected by
                                            the *axes* parameter of
                                            `filter_by_axis`

                ``filter_by_cell``          A sequence as expected by
                                            the *cells* parameter of
                                            `filter_by_cell`

                ``filter_by_connectivity``  A sequence as expected by
                                            the *connectivity*
                                            parameter of
                                            `filter_by_connectivities`

                ``filter_by_data``          Any value is allowed which
                                            will be ignored, as
                                            `filter_by_data` does not
                                            have any positional
                                            arguments.

                ``filter_by_identity``      A sequence as expected by
                                            the *identities* parameter
                                            of `filter_by_identity`

                ``filter_by_key``           A sequence as expected by
                                            the *keys* parameter of
                                            `filter_by_key`

                ``filter_by_measure``       A sequence as expected by
                                            the *measures* parameter
                                            of `filter_by_measure`

                ``filter_by_method``        A sequence as expected by
                                            the *methods* parameter of
                                            `filter_by_method`

                ``filter_by_naxes``         A sequence as expected by
                                            the *naxes* parameter of
                                            `filter_by_naxes`

                ``filter_by_ncdim``         A sequence as expected by
                                            the *ncdims* parameter of
                                            `filter_by_ncdim`

                ``filter_by_ncvar``         A sequence as expected by
                                            the *ncvars* parameter of
                                            `filter_by_ncvar`

                ``filter_by_property``      A dictionary as expected
                                            by the *properties*
                                            parameter of
                                            `filter_by_property`

                ``filter_by_size``          A sequence as expected by
                                            the *sizes* parameter of
                                            `filter_by_size`

                ``filter_by_type``          A sequence as expected by
                                            the *types* parameter of
                                            `filter_by_type`
                ==========================  ==========================

            axis_mode: `str`, optional
                Provide a value for the *axis_mode* parameter of the
                `filter_by_axis` method. By default *axis_mode* is
                ``'and'``.

            property_mode: `str`, optional
                Provide a value for the *property_mode* parameter of
                the `filter_by_property` method. By default
                *property_mode* is ``'and'``.

            {{todict: `bool`, optional}}

            {{cached: optional}}

            _identity_config: optional
                Provide a value for the *_config* parameter of the
                `filter_by_identity` method.

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        """
        if cached is not None:
            return cached

        if not filters:
            out, _ = self._filter_preprocess(self, todict=todict)
            return out

        out = self

        for method, args in filters.items():
            try:
                filter_method = getattr(self, "_" + method)
            except AttributeError:
                raise TypeError(
                    f"{self.__class__.__name__}.filter() has an unexpected "
                    f"keyword argument {method!r}"
                )

            args = self._filter_parse_args(
                method,
                args,
                todict=todict,
                axis_mode=axis_mode,
                property_mode=property_mode,
                _identity_config=_identity_config,
            )

            out = filter_method(out, *args)

        return out

    def _component_filter(
        self,
        arg,
        components,
        construct_types,
        filter_by,
        get_method,
        todict=False,
    ):
        """Worker function for some "_filter_by_*" methods.

        Used by `_filter_by_measure`, `_filter_by_method`,
        `_filter_by_cell`, and `_filter_by_connectivity` to filter by
        "measure", "method", "cell" and "connectivity" components,
        respectively.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            arg: `Constructs` or `dict`
                Either a `Constructs` object or its dictionary
                representation.

            components:
                Select *construct_type* constructs that have a
                component defined by their *get_method* methods, that
                matches any of the given values.

                If no components are provided then all
                *construct_type* constructs are selected.

                {{value match}}

            construct_types: sequence of `str`
                The construct types, e.g. ``['cell_measure']`` or
                ``['domain_topology', 'cell_connectivity']``.

            filter_by: `str`
                The name of the public "filter_by" method,
                e.g. ``'filter_by_measure'``.

            get_method: `str`
                The name of the construct method that gets the
                component, e.g. ``'get_measure'``.

            {{todict: `bool`, optional}}

        """
        out, pop = self._filter_preprocess(
            arg, filter_applied={filter_by: components}, todict=todict
        )

        all_construct_types = self._construct_type
        if not components:
            for cid in tuple(out):
                if all_construct_types[cid] not in construct_types:
                    pop(cid)

            return out

        for cid, construct in tuple(out.items()):
            if all_construct_types[cid] not in construct_types:
                pop(cid)
                continue

            value1 = getattr(construct, get_method)(None)

            ok = False
            if value1 is not None:
                for value0 in components:
                    ok = self._matching_values(
                        value0, None, value1, basic=True
                    )
                    if ok:
                        break

            if not ok:
                pop(cid)

        return out

    def _filter_by_axis(self, arg, axes, todict, axis_mode):
        """Worker function for `filter_by_axis` and `filter`.

        See `filter_by_axis` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        # Parse the axis_mode parameter
        _or = False
        _exact = False
        _subset = False
        if axis_mode in ("and", None):
            pass
        elif axis_mode == "or":
            _or = True
        elif axis_mode == "exact":
            _exact = True
        elif axis_mode == "subset":
            _subset = True
        elif axes:
            raise ValueError(
                f"{self.__class__.__name__}.filter_by_axis() has incorrect "
                f"'axis_mode' value {axis_mode!r}. "
                "Expected one of 'or', 'and', 'exact', 'subset'"
            )

        filter_applied = {"filter_by_axis": (axes, {"axis_mode": axis_mode})}

        if not axes:
            # Return all constructs that could have data if no axes
            # have been provided
            return self._filter_by_data(
                arg, None, todict, filter_applied=filter_applied
            )

        out, pop = self._filter_preprocess(
            arg, filter_applied=filter_applied, todict=todict
        )

        # Convert values to domain axis construct identifiers, if any
        # can be.
        axes2 = self._filter_convert_to_domain_axis(
            axes, check_axis_identities=True
        )

        if not axes2:
            # No arguments found unique domain axis constructs
            if isinstance(out, dict):
                out = {}
            else:
                out._clear()

            return out

        axes = set(axes2)

        data_axes = self._construct_axes

        for cid in tuple(out):
            x = data_axes.get(cid)
            if x is None:
                pop(cid)
                continue

            ok = True
            if _exact:
                if set(x) != axes:
                    ok = False
            elif _subset:
                if not set(x).issubset(axes):
                    ok = False
            else:
                for axis_key in axes:
                    ok = axis_key in x
                    if _or:
                        if ok:
                            break
                    elif not ok:
                        break

            if not ok:
                pop(cid)

        return out

    def filter_by_axis(
        self, *axes, axis_mode="and", todict=False, cached=None
    ):
        """Select metadata constructs by axes spanned by their data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            mode: `str`
                Deprecated at version 1.8.9.0. Use the *axis_mode*
                parameter instead.

            axes: optional
                Select constructs that whose data spans the domain
                axis constructs specified by the given values. A value
                may be:

                * A domain axis construct identifier, with or without
                  the ``'key%'`` prefix.

                * The unique domain axis construct spanned by all of
                  the 1-d coordinate constructs returned by, for a
                  given ``value``,
                  ``c.filter(filter_by_type=["dimension_coordinate",
                  "auxiliary_coordinate"], filter_by_naxes=(1,),
                  filter_by_identity=(value,))``. See `filter` for
                  details.

                * If there is an associated `Field` data array and a
                  value matches the integer position of an array
                  dimension, then the corresponding domain axis
                  construct is specified.

                * A unique domain axis construct identity, defined by
                  its `!identities` methods. In this case a value may
                  be any object that can match via the ``==``
                  operator, or a `re.Pattern` object that matches via
                  its `~re.Pattern.search` method.

                If no axes are provided then all constructs that do,
                or could have data, spanning any domain axes
                constructs, are selected.

            axis_mode: `str`
                Define the relationship between the given domain axes
                and the constructs' data.

                ===========  =========================================
                *axis_mode*  Description
                ===========  =========================================
                ``'and'``    A construct is selected if it spans *all*
                             of the given domain axes, *and possibly
                             others*.

                ``'or'``     A construct is selected if it spans *any*
                             of the domain axes, *and possibly
                             others*.

                ``exact``    A construct is selected if it spans *all*
                             of the given domain axes, *and no
                             others*.

                ``subset``   A construct is selected if it spans *a
                             subset* of the given domain axes, *and no
                             others*.
                ===========  =========================================

                By default *axis_mode* is ``'and'``.

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select constructs whose data spans the "domainaxis1" domain
        axis construct:

        >>> d = c.filter_by_axis('domainaxis1')

        Select constructs whose data does not span the "domainaxis2"
        domain axis construct:

        >>> d = c.filter_by_axis('domainaxis2').inverse_filter()

        Select constructs whose data spans the "domainaxis1", but not
        the "domainaxis2" domain axis constructs:

        >>> d = c.filter_by_axis('domainaxis1')
        >>> d = d.filter_by_axis('domainaxis2')
        >>> d  = d.inverse_filter(1)

        Select constructs whose data spans the "domainaxis1" or the
        "domainaxis2" domain axis constructs:

        >>> d = c.filter_by_axis('domainaxis1', 'domainaxis2', axis_mode="or")

        """
        if cached is not None:
            return cached

        return self._filter_by_axis(self, axes, todict, axis_mode)

    def _filter_by_cell(self, arg, cells, todict):
        """Worker function for `filter_by_cell` and `filter`.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self._component_filter(
            arg,
            cells,
            ("domain_topology",),
            "filter_by_cell",
            "get_cell",
            todict,
        )

    def filter_by_cell(self, *cells, todict=False, cached=None):
        """Select domain topology constructs by cell type.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            cells: optional
                Select domain topology constructs that have a cell
                type, defined by their `!get_cell` methods, that
                matches any of the given values.

                If no cell types are provided then all domain topology
                constructs are selected.

                {{value match}}

            {{todict: `bool`, optional}}

            {{cached: optional}}

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected domain topology constructs, or a cached
                valued.

        **Examples**

        >>> print(t.constructs.filter_by_type('domain_topology')
        Constructs:
        {'domaintopology0': <{{repr}}DomainTopology: cell:face>}

        Select domain topology constructs that have a cell type of
        "face":

        >>> print(c.filter_by_cell('face')
        Constructs:
        {'domaintopology0': <{{repr}}DomainTopology: cell:face>}

        Select domain topology constructs that have a cell type of
        'face' or 'edge':

        >>> print(c.filter_by_cell('face', 'edge')
        Constructs:
        {'domaintopology0': <{{repr}}DomainTopology: cell:face>}

        Select domain topology constructs that have a cell type that
        ends with "e":

        >>> print(c.filter_by_cell(re.compile('e$')))
        Constructs:
        {'domaintopology0': <{{repr}}DomainTopology: cell:face>}

        Select domain topology constructs that have a cell type of any
        value:

        >>> print(c.filter_by_cell())
        Constructs:
        {'domaintopology0': <{{repr}}DomainTopology: cell:face>}

        """
        if cached is not None:
            return cached

        return self._filter_by_cell(self, cells, todict)

    def _filter_by_connectivity(self, arg, connectivities, todict):
        """Worker function for `filter_by_connectivity` and `filter`.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self._component_filter(
            arg,
            connectivities,
            ("cell_connectivity",),
            "filter_by_connectivity",
            "get_connectivity",
            todict,
        )

    def filter_by_connectivity(
        self, *connectivities, todict=False, cached=None
    ):
        """Select cell connectivity constructs by connectivity type.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            connectivities: optional
                Select cell connectivity constructs that have a
                connectivity type, defined by their `!get_connectivity`
                methods, that matches any of the given values.

                If no connectivities are provided then all cell
                connectivity constructs are selected.

                {{value match}}

            {{todict: `bool`, optional}}

            {{cached: optional}}

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected cell connectivity constructs, or a cached
                value.

        **Examples**

        >>> print(t.constructs.filter_by_type('cell_connectivity')
        Constructs:
        {'cellconnectivity0': <{{repr}}CellConnectivity: connectivity:edge(13824) >,
         'cellconnectivity1': <{{repr}}CellConnectivity: connectivity:node(13824) >}

        Select cell connectivity constructs that have a connectivity
        type of "edge":

        >>> print(c.filter_by_connectivity('edge')
        Constructs:
        {'cellconnectivity0': <{{repr}}CellConnectivity: connectivity:edge(13824) >,

        Select cell connectivity constructs that have a connectivity
        type of 'edge' or 'node':

        >>> print(c.filter_by_connectivity('edge', 'node')
        Constructs:
        {'cellconnectivity0': <{{repr}}CellConnectivity: connectivity:edge(13824) >,
         'cellconnectivity1': <{{repr}}CellConnectivity: connectivity:node(13824) >}

        Select cell connectivity constructs that have a connectivity
        type that starts with "n":

        >>> print(c.filter_by_connectivity(re.compile('^n')))
        Constructs:
        {'cellconnectivity1': <{{repr}}CellConnectivity: connectivity:node(13824) >}

        Select cell connectivity constructs that have a connectivity
        type of any value:

        >>> print(c.filter_by_connectivity())
        Constructs:
        {'cellconnectivity0': <{{repr}}CellConnectivity: connectivity:edge(13824) >,
         'cellconnectivity1': <{{repr}}CellConnectivity: connectivity:node(13824) >}

        """
        if cached is not None:
            return cached

        return self._filter_by_connectivity(self, connectivities, todict)

    def _filter_by_data(self, arg, ignored, todict, filter_applied=None):
        """Worker function for `filter_by_data` and `filter`.

        See `filter_by_data` for details.

        :Parameters:

            ignored:
                This paramaeter is always ignored, but needs to be set
                to something in order to satisfy with the `filter`
                API.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        return self._filter_by_type(
            arg, self._array_constructs, todict, filter_applied=filter_applied
        )

    def filter_by_data(self, todict=False, cached=None):
        """Selects metadata constructs that could contain data.

        Selection is not based on whether they actually have data,
        rather by whether the construct supports the inclusion of
        data. For example, constructs selected by this method will all
        have a `!get_data` method.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select constructs that could contain data:

        >>> d = c.filter_by_data()

        """
        if cached is not None:
            return cached

        return self._filter_by_data(self, None, todict, filter_applied=None)

    def _filter_by_identity(self, arg, identities, todict, _config):
        """Worker function for `filter_by_identity` and `filter`.

        See `filter_by_identity` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        out, pop = self._filter_preprocess(
            arg,
            filter_applied={"filter_by_identity": identities},
            todict=todict,
        )

        if not identities:
            # Return all constructs if no identities have been provided
            return out

        # ------------------------------------------------------------
        # The principle here is to iterate over as few individual
        # construct identities as possible, as these can be very
        # expensive to compute.
        # ------------------------------------------------------------
        return_matched = _config.get("return_matched")
        short_iteration = _config.get("short_iteration", self._short_iteration)
        identities_kwargs = _config.get("identities_kwargs", {})

        matched = set()
        hits = []

        # Process identities that are construct identifiers
        identities2 = []
        for value0 in identities:
            if isinstance(value0, str):
                if value0 in out:
                    matched.add(value0)
                    hits.append(value0)
                    continue
                elif value0.startswith("key%") and value0[4:] in out:
                    matched.add(value0[4:])
                    hits.append(value0)
                    continue

            identities2.append(value0)

        if identities2:
            if matched:
                constructs = {
                    cid: c for cid, c in out.items() if cid not in matched
                }
            else:
                constructs = out

            if "short" not in identities_kwargs:
                short = True
                for value0 in identities:
                    if not short_iteration(value0):
                        short = False
                        break

                identities_kwargs["short"] = short

            # Dictionary of construct identifiers and construct
            # identity generators
            generators = {
                cid: construct.identities(generator=True, **identities_kwargs)
                for cid, construct in constructs.items()
            }

            for values in zip_longest(*generators.values(), fillvalue=None):
                # Loop round the each construct's next identity
                for (cid, generator), value1 in zip(
                    generators.items(), values
                ):
                    if value1 is None:
                        # This construct has run out of identities
                        continue

                    # Loop round the given values
                    for value0 in identities:
                        if self._matching_values(
                            value0, None, value1, basic=True
                        ):
                            hits.append(value0)
                            matched.add(cid)
                            break

        if return_matched:
            hits = set(hits)
            misses = hits.symmetric_difference(identities)
        else:
            misses = False

        if not misses:
            if not matched:
                if isinstance(out, dict):
                    out = {}
                else:
                    out._clear()
            elif len(matched) <= len(out):
                for cid in tuple(out):
                    if cid not in matched:
                        pop(cid)

        if return_matched:
            if misses:
                out = None

            return out, matched, hits, misses

        return out

    def filter_by_identity(
        self, *identities, todict=False, cached=None, _config={}
    ):
        """Select metadata constructs by identity.

        Calling a `Constructs` instance is an alias for
        `filter_by_identity`.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            identities: optional
                Select constructs that have an identity, defined by
                their `!identities` methods, that matches any of the
                given values.

            * A metadata construct identity.

              {{construct selection identity}}

            * The key of a metadata construct

            *Parameter example:*
              ``identity='latitude'``

                {{value match}}

                {{displayed identity}}

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            _config: optional
                Additional parameters for configuring the application
                of a construct's `identities` method.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select constructs that have a ``standard_name`` property of
        'latitude':

        >>> d = c.filter_by_identity('latitude')

        Select constructs that have a ``long_name`` property of
        'Height':

        >>> d = c.filter_by_identity('long_name=Height')

        Select constructs that have a ``standard_name`` property of
        'latitude' or a "foo" property of 'bar':

        >>> d = c.filter_by_identity('latitude', 'foo=bar')

        Select constructs that have a netCDF variable name of 'time':

        >>> d = c.filter_by_identity('ncvar%time')

        """
        if cached is not None:
            return cached

        return self._filter_by_identity(self, identities, todict, _config)

    def _filter_by_key(self, arg, keys, todict):
        """Worker function for `filter_by_key` and `filter`.

        See `filter_by_key` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        out, pop = self._filter_preprocess(
            arg, filter_applied={"filter_by_key": keys}, todict=todict
        )

        if not keys:
            return out

        for cid in tuple(out):
            if cid in keys:
                continue

            ok = False
            for value0 in keys:
                ok = self._matching_values(value0, None, cid, basic=True)
                if ok:
                    break

            if not ok:
                pop(cid)

        return out

    def filter_by_key(self, *keys, todict=False, cached=None):
        """Select metadata constructs by key.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            key: optional
                Select constructs that have a construct identifier
                (e.g. ``'dimensioncoordinate1'``) that matches any of
                the given values.

                If no keys are provided then all constructs are
                selected.

                {{value match}}

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select the construct with key 'domainancillary0':

        >>> d = c.filter_by_key('domainancillary0')

        Select the constructs with keys 'dimensioncoordinate1' or
        'fieldancillary0':

        >>> d = c.filter_by_key('dimensioncoordinate1', 'fieldancillary0')

        """
        if cached is not None:
            return cached

        return self._filter_by_key(self, keys, todict)

    def _filter_by_measure(self, arg, measures, todict):
        """Worker function for `filter_by_measure` and `filter`.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        return self._component_filter(
            arg,
            measures,
            ("cell_measure",),
            "filter_by_measure",
            "get_measure",
            todict,
        )

    def filter_by_measure(self, *measures, todict=False, cached=None):
        """Select cell measure constructs by measure.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            measures: optional
                Select cell measure constructs that have a measure,
                defined by their `!get_measure` methods, that matches
                any of the given values.

                If no measures are provided then all cell measure
                constructs are selected.

                {{value match}}

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

            `Constructs`
                The selected cell measure constructs and their
                construct keys.

        **Examples**

        >>> print(t.constructs.filter_by_type('measure'))
        Constructs:
        {'cellmeasure0': <{{repr}}CellMeasure: measure:area(9, 10) km2>,
         'cellmeasure1': <{{repr}}CellMeasure: measure:volume(3, 9, 10) m3>}

        Select cell measure constructs that have a measure of 'area':

        >>> print(c.filter_by_measure('area'))
        Constructs:
        {'cellmeasure0': <{{repr}}CellMeasure: measure:area(9, 10) km2>}

        Select cell measure constructs that have a measure of 'area'
        or 'volume':

        >>> print(c.filter_by_measure('area', 'volume'))
        Constructs:
        {'cellmeasure0': <{{repr}}CellMeasure: measure:area(9, 10) km2>,
         'cellmeasure1': <{{repr}}CellMeasure: measure:volume(3, 9, 10) m3>}

        Select cell measure constructs that have a measure of start
        with the letter "a" or "v":

        >>> print(c.filter_by_measure(re.compile('^a|v')))
        Constructs:
        {'cellmeasure0': <{{repr}}CellMeasure: measure:area(9, 10) km2>,
         'cellmeasure1': <{{repr}}CellMeasure: measure:volume(3, 9, 10) m3>}

        Select cell measure constructs that have a measure of any
        value:

        >>> print(c.filer_by_measure())
        Constructs:
        {'cellmeasure0': <{{repr}}CellMeasure: measure:area(9, 10) km2>,
         'cellmeasure1': <{{repr}}CellMeasure: measure:volume(3, 9, 10) m3>}

        """
        if cached is not None:
            return cached

        return self._filter_by_measure(self, measures, todict)

    def _filter_by_method(self, arg, methods, todict):
        """Worker function for `filter_by_method` and `filter`.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        return self._component_filter(
            arg,
            methods,
            ("cell_method",),
            "filter_by_method",
            "get_method",
            todict,
        )

    def filter_by_method(self, *methods, todict=False, cached=None):
        """Select cell method constructs by method.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            methods: optional
                Select cell method constructs that have a method,
                defined by their `!get_method` methods, that matches
                any of the given values.

                If no methods are provided then all cell method
                constructs are selected.

                {{value match}}

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        >>> print(c.constructs.filter_by_type('cell_method'))
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
         'cellmethod1': <{{repr}}CellMethod: domainaxis3: maximum>}

        Select cell method constructs that have a method of 'mean':

        >>> print(c.filter_by_method('mean'))
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>}

        Select cell method constructs that have a method of 'mean' or
        'maximum':

        >>> print(c.filter_by_method('mean', 'maximum'))
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
         'cellmethod1': <{{repr}}CellMethod: domainaxis3: maximum>}

        Select cell method constructs that have a method that contain
        the letter 'x':

        >>> import re
        >>> print(c.filter_by_method(re.compile('x')))
        Constructs:
        {'cellmethod1': <{{repr}}CellMethod: domainaxis3: maximum>}

        Select cell method constructs that have a method of any value:

        >>> print(c.filter_by_method())
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
         'cellmethod1': <{{repr}}CellMethod: domainaxis3: maximum>}

        """
        if cached is not None:
            return cached

        return self._filter_by_method(self, methods, todict)

    def _filter_by_naxes(self, arg, naxes, todict):
        """Worker function for `filter_by_naxes` and `filter`.

        See `filter_by_naxes` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        if not naxes:
            # If no naxes have been provided then return all
            # constructs that could have data
            return self._filter_by_data(
                arg, None, todict, filter_applied={"filter_by_naxes": naxes}
            )

        out, pop = self._filter_preprocess(
            arg, filter_applied={"filter_by_naxes": naxes}, todict=todict
        )

        data_axes = self._construct_axes

        for key in tuple(out):
            x = data_axes.get(key)
            if x is None:
                pop(key)
                continue

            ok = len(x) in naxes
            if not ok:
                pop(key)

        return out

    def filter_by_naxes(self, *naxes, todict=False, cached=None):
        """Selects constructs by the number of axes their data spans.

        Specifically, selects metadata constructs by the number of
        domain axis constructs spanned by their data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            naxes: optional
                Select constructs that have data whose number of
                dimensions matches any of the given values.

                If no values are provided then all constructs that do
                or could have data, spanning any domain axes
                constructs, are selected.

                A value may be any object that can match an integer
                via ``==``.

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select constructs that contain data that spans two domain axis
        constructs:

        >>> d = c.filter_by_naxes(2)

        Select constructs that contain data that spans one or two
        domain axis constructs:

        >>> d = c.filter_by_ncdim(1, 2)

        """
        if cached is not None:
            return cached

        return self._filter_by_naxes(self, naxes, todict)

    def _filter_by_ncdim(self, arg, ncdims, todict):
        """Worker function for `filter_by_ncdim` and `filter`.

        See `filter_by_ncdim` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        out, pop = self._filter_preprocess(
            arg, filter_applied={"filter_by_ncdim": ncdims}, todict=todict
        )

        if not ncdims:
            # If no ncdims have been provided then return all
            # constructs that could have a netCDF dimension name
            for cid, construct in tuple(out.items()):
                try:
                    construct.nc_get_dimension
                except AttributeError:
                    pop(cid)

            return out

        for cid, construct in tuple(out.items()):
            try:
                nc_get_dimension = construct.nc_get_dimension
            except AttributeError:
                pop(cid)
                continue

            value1 = nc_get_dimension(None)

            ok = False
            if value1 is not None:
                for value0 in ncdims:
                    ok = self._matching_values(
                        value0, None, value1, basic=True
                    )
                    if ok:
                        break

            if not ok:
                pop(cid)

        return out

    def filter_by_ncdim(self, *ncdims, todict=False, cached=None):
        """Select domain axis constructs by netCDF dimension name.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            ncdims: optional
                Select constructs that have a netCDF dimension name,
                defined by their `!nc_get_dimension` methods, that
                match any of the given values.

                If no netCDF dimension names are provided then all
                constructs that do or could have a netCDF dimension
                name are selected.

                {{value match}}

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select the domain axis constructs with netCDF dimension name
        'time':

        >>> d = c.filter_by_ncdim('time')

        Select the domain axis constructs with netCDF dimension name
        'time' or 'lat':

        >>> d = c.filter_by_ncdim('time', 'lat')

        """
        if cached is not None:
            return cached

        return self._filter_by_ncdim(self, ncdims, todict)

    def _filter_by_ncvar(self, arg, ncvars, todict):
        """Worker function for `filter_by_ncvar` and `filter`.

        See `filter_by_ncvar` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        out, pop = self._filter_preprocess(
            arg, filter_applied={"filter_by_ncvar": ncvars}, todict=todict
        )

        if not ncvars:
            # If no ncvars have been provided then return all
            # constructs that could have a netCDF variable name
            for cid, construct in tuple(out.items()):
                try:
                    construct.nc_get_variable
                except AttributeError:
                    pop(cid)

            return out

        for cid, construct in tuple(out.items()):
            try:
                nc_get_variable = construct.nc_get_variable
            except AttributeError:
                pop(cid)
                continue

            value1 = nc_get_variable(None)

            ok = False
            if value1 is not None:
                for value0 in ncvars:
                    ok = self._matching_values(
                        value0, None, value1, basic=True
                    )
                    if ok:
                        break

            if not ok:
                pop(cid)

        return out

    def filter_by_ncvar(self, *ncvars, todict=False, cached=None):
        """Select domain axis constructs by netCDF variable name.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            ncvars: optional
                Select constructs that have a netCDF variable name,
                defined by their `!nc_get_variable` methods, that
                match any of the given values.

                If no netCDF variable names are provided then all
                constructs that do or could have a netCDF variable
                name are selected.

                {{value match}}

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select the constructs with netCDF variable name 'time':

        >>> d = c.filter_by_ncvar('time')

        Select the constructs with netCDF variable name 'time' or
        'lat':

        >>> d = c.filter_by_ncvar('time', 'lat')

        """
        if cached is not None:
            return cached

        return self._filter_by_ncvar(self, ncvars, todict)

    def _filter_by_property(self, arg, properties, todict, property_mode):
        """Worker function for `filter_by_property` and `filter`.

        See `filter_by_property`  for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        # Parse property_mode
        if not property_mode:
            # property_mode is 'and' by default
            _or = False
        else:
            if len(property_mode) > 1:
                raise ValueError(
                    f"{self.__class__.__name__}.filter_by_property() accepts"
                    "at most one positional parameter, "
                    f"got {len(property_mode)}"
                )

            mode = property_mode[0]
            if mode == "or":
                _or = True
            elif mode == "and":
                _or = False
            else:
                raise ValueError(
                    f"{self.__class__.__name__}.filter_by_property() has "
                    f"incorrect 'property_mode' value ({mode!r}). "
                    "If set must be one of 'or', 'and'"
                )

        out, pop = self._filter_preprocess(
            arg,
            filter_applied={"filter_by_property": (property_mode, properties)},
            todict=todict,
        )

        if not properties:
            for cid, construct in tuple(out.items()):
                try:
                    construct.get_property
                except AttributeError:
                    pop(cid)

            return out

        for cid, construct in tuple(out.items()):
            try:
                get_property = construct.get_property
            except AttributeError:
                pop(cid)
                continue

            ok = True
            for name, value0 in properties.items():
                value1 = get_property(name, None)

                if value1 is None:
                    ok = False
                elif value0 is None:
                    # If a given value is None then match if the
                    # construct the property, regardless of the
                    # construct's property value.
                    ok = True
                else:
                    ok = self._matching_values(value0, construct, value1)

                if _or:
                    if ok:
                        break
                elif not ok:
                    break

            if not ok:
                pop(cid)

        return out

    def filter_by_property(self, *property_mode, **properties):
        """Select metadata constructs by property.

        Unlike the other "filter_by_" methods, this method has no
        *todict* or *cached* parameters. To get the results as a
        dictionary, use the `todict` method on the result, or for
        faster performance use the `filter` method. For instance, to
        get a dictionary of all constructs with a long_name of
        ``'time'`` you could do
        ``c.filter_by_property(long_name="time").todict()``, or the
        faster ``d = c.filter(filter_by_property={"long_name":
        "time"}, todict=True)``. To return a cached result, you have
        to use the `filter` method:
        ``c.filter(filter_by_property={"long_name": "time"},
        cached=x)``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            property_mode: optional
                Define the behaviour when multiple properties are
                provided:

                ===============  =====================================
                *property_mode*  Description
                ===============  =====================================
                ``'and'``        A construct is selected if it matches
                                 all of the given properties.

                ``'or'``         A construct is selected when at least
                                 one of its properties matches.
                ===============  =====================================

                By default *property_mode* is ``'and'``.

            properties:  optional
                Select constructs that have a CF property, defined by
                their `!properties` methods, that matches any of the
                given values.

                A property names and their given values are specified
                by keyword argument names and values.

                {{value match}}

                The special value of `None` selects any construct that
                has that property, regardless of the construct's
                property value.

                If no properties are provided then all constructs that
                do or could have CF properties, with any values, are
                selected.

        :Returns:

            `Constructs`
                The selected constructs.

        **Examples**

        Select constructs that have a ``standard_name`` of 'latitude':

        >>> d = c.filter_by_property(standard_name='latitude')

        Select constructs that have a ``long_name`` of 'height' *and*
        ``units`` of 'm':

        >>> d = c.filter_by_property(long_name='height', units='m')

        Select constructs that have a ``long_name`` of 'height' *or* a
        ``foo`` of 'bar':

        >>> d = c.filter_by_property('or', long_name='height', foo='bar')

        Select constructs that have a ``standard_name`` which contains
        start with the string 'air':

        >>> import re
        >>> d = c.filter_by_property(standard_name=re.compile('^air'))

        """
        return self._filter_by_property(self, properties, False, property_mode)

    def _filter_by_size(self, arg, sizes, todict):
        """Worker function for `filter_by_size` and `filter`.

        See `filter_by_size` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        if not sizes:
            # If no sizes have been provided then return all domain
            # axis constructs
            return self._filter_by_type(
                arg,
                ("domain_axis",),
                todict,
                filter_applied={"filter_by_size": sizes},
            )

        out, pop = self._filter_preprocess(
            arg, filter_applied={"filter_by_size": sizes}, todict=todict
        )

        construct_type = self._construct_type
        for cid, construct in tuple(out.items()):
            if construct_type[cid] != "domain_axis":
                pop(cid)
                continue

            value1 = construct.get_size(None)

            ok = value1 is not None and value1 in sizes
            if not ok:
                pop(cid)

        return out

    def filter_by_size(self, *sizes, todict=False, cached=None):
        """Select domain axis constructs by size.

        .. versionadded:: (cfdm) 1.7.3

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            sizes: optional
                Select domain axis constructs that have sizes, defined
                by their `!get_size` methods, that match any of the
                given values.

                If no sizes are provided then all domain axis
                constructs are selected.

                A value may be any object that can match an integer
                via ``==``.

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select domain axis constructs that have a size of 1:

        >>> d = c.filter_by_size(1)

        Select domain axis constructs that have a size of 1 or 96:

        >>> d = c.filter_by_size(1, 96)

        """
        if cached is not None:
            return cached

        return self._filter_by_size(self, sizes, todict)

    def _filter_by_type(self, arg, types, todict, filter_applied=None):
        """Worker function for `filter_by_type` and `filter`.

        See `filter_by_type` for details.

        .. versionadded:: (cfdm) 1.8.9.0

        """
        if isinstance(arg, dict):
            if types:
                pop = arg.pop
                construct_type = self._construct_type
                for cid in tuple(arg):
                    if construct_type[cid] not in types:
                        pop(cid)

            return arg

        out = super(Constructs, arg).filter_by_type(*types, todict=todict)

        if not todict:
            if filter_applied is None:
                filter_applied = {"filter_by_type": types}

            out._prefiltered = self.shallow_copy()
            out._filters_applied = arg.filters_applied() + (filter_applied,)

        return out

    def filter_by_type(self, *types, todict=False, cached=None):
        """Select metadata constructs by type.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

            types: optional
                Select constructs that have are of any of the given
                types.

                A type is specified by one of the following strings:

                ==========================  ================================
                *type*                      Construct selected
                ==========================  ================================
                ``'domain_ancillary'``      Domain ancillary constructs
                ``'dimension_coordinate'``  Dimension coordinate constructs
                ``'domain_axis'``           Domain axis constructs
                ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
                ``'cell_measure'``          Cell measure constructs
                ``'coordinate_reference'``  Coordinate reference constructs
                ``'domain_topology'``       Domain topology constructs
                ``'cell_connectivity'``     Cell connectivity constructs
                ``'cell_method'``           Cell method constructs
                ``'field_ancillary'``       Field ancillary constructs
                ==========================  ================================

                If no types are provided then all constructs are selected.

            {{todict: `bool`, optional}}

                .. versionadded:: (cfdm) 1.8.9.0

            {{cached: optional}}

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `Constructs` or `dict` or *cached*
                The selected constructs, or a cached valued.

        **Examples**

        Select dimension coordinate constructs:

        >>> d = c.filter_by_type('dimension_coordinate')

        Select dimension coordinate and field ancillary constructs:

        >>> d = c.filter_by_type('dimension_coordinate', 'field_ancillary')

        """
        if cached is not None:
            return cached

        return self._filter_by_type(self, types, todict)

    def filters_applied(self):
        """A history of filters that have been applied.

        The history is returned in a tuple. The last element of the
        tuple describes the last filter applied. Each element is a
        single-entry dictionary whose key is the name of the filter
        method that was used, with a value that gives the arguments
        that were passed to the call of that method. If no filters
        have been applied then the tuple is empty.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `inverse_filter`,
                     `clear_filters_applied`, `unfilter`

        :Returns:

            `tuple`
                The history of filters that have been applied, ordered
                from first to last. If no filters have been applied
                then the tuple is empty.

        **Examples**

        >>> print(c)
        {'auxiliarycoordinate0': <{{repr}}AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
         'auxiliarycoordinate1': <{{repr}}AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
         'coordinatereference1': <{{repr}}CoordinateReference: grid_mapping_name:rotated_latitude_longitude>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: grid_latitude(10) degrees>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: grid_longitude(9) degrees>,
         'domainaxis1': <{{repr}}DomainAxis: size(10)>,
         'domainaxis2': <{{repr}}DomainAxis: size(9)>}
        >>> c.filters_applied()
        ()
        >>> c = c.filter_by_type('dimension_coordinate', 'auxiliary_coordinate')
        >>> c.filters_applied()
        ({'filter_by_type': ('dimension_coordinate', 'auxiliary_coordinate')},)
        >>> c = c.filter_by_property(units='degrees')
        >>> c.filters_applied()
        ({'filter_by_type': ('dimension_coordinate', 'auxiliary_coordinate')},
         {'filter_by_property': ((), {'units': 'degrees'})})
        >>> c = c.filter_by_property('or', standard_name='grid_latitude', axis='Y')
        >>> c.filters_applied()
        ({'filter_by_type': ('dimension_coordinate', 'auxiliary_coordinate')},
         {'filter_by_property': ((), {'units': 'degrees'})},
         {'filter_by_property': (('or',), {'axis': 'Y', 'standard_name': 'grid_latitude'})})
        >>> print(c)
        Constructs:
        {'dimensioncoordinate1': <{{repr}}DimensionCoordinate: grid_latitude(10) degrees>}

        """
        filters = getattr(self, "_filters_applied", None)
        if filters is None:
            return ()

        return deepcopy(filters)

    def clear_filters_applied(self):
        """Remove the history of filters that have been applied.

        This method does not change the metadata constructs, it just
        forgets the history of any filters that have previously been
        applied. Use `inverse_filter` or `unfilter` retrieve
        previously filtered constructs.

        The removed history is returned in a tuple. The last element
        of the tuple describes the last filter applied. Each element
        is a single-entry dictionary whose key is the name of the
        filter method that was used, with a value that gives the
        arguments that were passed to the call of that method. If no
        filters have been applied then the tuple is empty.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`, `inverse_filter`,
                     `unfilter`

        :Returns:

            `tuple`
                The removed history of filters that have been applied,
                ordered from first to last. If no filters have been
                applied then the tuple is empty.


        **Examples**

        >>> c.filters_applied()
        ({'filter_by_naxes': (3, 1)},
         {'filter_by_identity': ('grid_longitude',)})
        >>> c.clear_filters_applied()
        ({'filter_by_naxes': (3, 1)},
         {'filter_by_identity': ('grid_longitude',)})
        >>> c.filters_applied()
        ()

        """
        out = self.filters_applied()
        self._filters_applied = None
        self._prefiltered = None
        return out

    def inverse_filter(self, depth=None):
        """Return the inverse of previous filters.

        By default, the inverse comprises all of the constructs that
        were *not* selected by all previously applied filters. If no
        filters have been applied, then this will result in empty
        `Constructs` instance being returned.

        If the *depth* parameter is set to *N* then the inverse is
        relative to the constructs selected by the *N*-th most
        recently applied filter.

        A history of the filters that have been applied is returned in
        a `tuple` by the `filters_applied` method. The last element of
        the tuple describes the last filter applied. If no filters
        have been applied then the tuple is empty.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`,
                     `clear_filters_applied`, `unfilter`

        :Parameters:

             depth: `int`, optional
                If set to ``N`` then the inverse is relative to the
                constructs selected by the ``N``-th most recently
                applied filter. By default the inverse is relative to
                the constructs selected by all previously applied
                filters. ``N`` may be larger than the total number of
                filters applied, which results in the default
                behaviour.

        :Returns:

            `Constructs`
                The constructs, and their construct keys, that were
                not selected by the last filter applied. If no
                filtering has been applied, or the last filter was an
                inverse filter, then an empty `Constructs` instance is
                returned.

        **Examples**

        >>> print(c)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}
        >>> print(c.inverse_filter())
        Constructs:
        {}
        >>> d = c.filter_by_type('dimension_coordinate', 'cell_method')
        >>> print(d)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >}
        >>> print(d.inverse_filter())
        Constructs:
        {'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}
        >>> e = d.filter_by_method('mean')
        >>> print(e)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>}
        >>> print(e.inverse_filter(1))
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >}
        >>> print(e.inverse_filter())
        Constructs:
        {'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}
        >>> print(e.inverse_filter(1).inverse_filter())
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}

        """
        out = self.unfilter(depth=depth)

        if depth:
            if "inverse_filter" in self.filters_applied()[-1]:
                filters = out.filters_applied()
                d = 1
                while True:
                    if d > len(filters) or "inverse_filter" not in filters[-d]:
                        break

                    d += 1

                if d > 1:
                    out = self.unfilter(depth=depth + d - 1)

                return out

        for key in self:
            out._pop(key)

        out._filters_applied = self.filters_applied() + (
            {"inverse_filter": ()},
        )

        out._prefiltered = self.shallow_copy()

        return out

    def shallow_copy(self, _ignore=None):
        """Return a shallow copy.

        ``f.shallow_copy()`` is equivalent to ``copy.copy(f)``.

        Any in-place changes to the actual constructs of the copy will
        not be seen in the original `{{class}}` object, and vice
        versa, but the copy is otherwise independent of its parent.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `Constructs`
                The shallow copy.

        **Examples**

        >>> g = f.shallow_copy()

        """
        out = super().shallow_copy(_ignore=_ignore)

        prefiltered = getattr(self, "_prefiltered", None)
        if prefiltered is not None:
            out._prefiltered = prefiltered.shallow_copy()
            out._filters_applied = self._filters_applied

        return out

    def unfilter(self, depth=None, copy=True):
        """Return the constructs that existed prior to previous filters.

        By default, the unfiltered constructs are those that existed
        before all previously applied filters.

        If the *depth* parameter is set to *N* then the unfiltered
        constructs are those that existed before the *N*-th most
        recently applied filter.

        A history of the filters that have been applied is returned in
        a `tuple` by the `filters_applied` method. The last element of
        the tuple describes the last filter applied. If no filters
        have been applied then the tuple is empty.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `filter`, `filters_applied`,
                     `clear_filters_applied`, `inverse_filter`

        :Parameters:

             depth: `int`, optional
                If set to ``N`` then return the constructs selected by
                the ``N``-th most recently applied filter. By default
                the constructs from before all previously applied
                filters are returned. ``N`` may be larger than the
                total number of filters applied, which results in the
                default behaviour.

        :Returns:

            `Constructs`
                The constructs, and their construct keys, that existed
                before the last filter was applied. If no filters have
                been applied then all of the constructs are returned.

        **Examples**

        >>> print(c)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}
        >>> d = c.filter_by_type('dimension_coordinate', 'cell_method')
        >>> print(d)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >}
        >>> d.unfilter().equals(c)
        True
        >>> e = d.filter_by_method('mean')
        >>> print(e)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>}
        >>> c.unfilter().equals(c)
        True
        >>> c.unfilter(0).equals(c)
        True
        >>> e.unfilter().equals(c)
        True
        >>> e.unfilter(1).equals(d)
        True
        >>> e.unfilter(2).equals(c)
        True

        If no filters have been applied then the unfiltered constructs are
        unchanged:

        >>> c.filters_applied()
        ()
        >>> c.unfilter().equals(c)
        True

        """
        out = self

        if depth is None:
            while True:
                prefiltered = getattr(out, "_prefiltered", None)
                if prefiltered is None:
                    break

                out = prefiltered
        else:
            for _ in range(depth):
                prefiltered = getattr(out, "_prefiltered", None)
                if prefiltered is not None:
                    out = prefiltered
                else:
                    break

        if copy:
            out = out.shallow_copy()

        return out



import numpy as np

from .quadraticinterpolation import QuadraticInterpolation

x, y, z = (0, 1, 2)


class QuadraticGeographicInterpolation(QuadraticInterpolation):
    """Mixin class for quadratic geographic interpolation.

    See CF appendix J "Coordinate Interpolation Methods".

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def _fcea2cv(self, va, vb, ce, ca):
        """The three-dimensional cartesian interpolation coefficients.

        The three-dimensional cartesian representation of the
        interpolation coefficients calculated from the parametric
        representation.

        cv = fcea2cv(va, vb, ce, ca)
           = fplus(fmultiply(ce, fminus(va, vb)),
                   fmultiply(ca, fcross(va, vb)),
                   fmultiply(cr, vr))

        where

        vr = fmultiply(0.5, fplus(va, vb))
        rsqr = fdot(vr, vr)
        cr = fsqrt(1 - ce*ce - ca*ca) - fsqrt(rsqr)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            va: `tuple` of `numpy.ndarray`

            vb: `tuple` of `numpy.ndarray`

            ce: `numpy.ndarray`

            ca: `numpy.ndarray`

        :Returns:

            `numpy.ndarray`

        """
        vr = self._fmultiply(0.5, self._fplus(va, vb))
        rsqr = self._fdot(vr, vr)

        k = 1
        if ce is not None:
            k = k - ce * ce

        if ca is not None:
            k = k - ca * ca

        cr = self._fsqrt(k) - self._fsqrt(rsqr)

        cv = self._fmultiply(cr, vr)

        if ce is not None:
            cv = self._fplus(cv, self._fmultiply(ce, self._fminus(va, vb)))

        if ca is not None:
            cv = self._fplus(cv, self._fmultiply(ca, self._fcross(va, vb)))

        return cv

    def _fcross(self, va, vb):
        """Vector cross product.

        (x, y, z) = fcross(va, vb)
                  = (va.y*vb.z - va.z*vb.y,
                     va.z*vb.x - va.x*vb.z,
                     va.x*vb.y - va.y*vb.x)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            va: `tuple` of `numpy.ndarray`

            vb: `tuple` of `numpy.ndarray`

        :Returns:

            `tuple` of `numpy.ndarray`

        """
        return (
            va[y] * vb[z] - va[z] * vb[y],
            va[z] * vb[x] - va[x] * vb[z],
            va[x] * vb[y] - va[y] * vb[x],
        )

    def _fcv(self, va, vb, vp_i, d, s_i):
        """Three-dimensionsal cartesian interpolation coefficients.

        cv = fcv(va, vb, vp(i), s(i))
           = (fw(va.x, vb.x, vp(i).x, s(i)),
              fw(va.y, vb.y, vp(i).y, s(i)),
              fw(va.z, vb.z, vp(i).z, s(i)))

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            va, vb: `tuple` of `numpy.ndarray`
                The vector representation of the tie point (latitude,
                longitude).

            vp_i: `tuple` of `numpy.ndarray`
                The vector representation of the uncompressed value at
                the same location as *s_i*.

            {{d: `int`}}

            {{s_i: array_like}}

        :Returns:

            `tuple` of `numpy.ndarray`

        """
        return tuple(
            self._fw(va[i], vb[i], vp_i[i], d, s_i) for i in (x, y, z)
        )

    def _fcv2cea(self, va, vb, cv):
        """Parametric representation interpolation coefficients.

        ce, ca = fcv2cea(va, vb, cv)
               = (fdot(cv, fminus(va, vb)) / gsqr,
                  fdot(cv, fcross(va, vb)) / (rsqr*gsqr))

        where

        rsqr = fdot(vr, vr)
        gsqr = fdot(vg, vg)

        where

        vr = fmultiply(0.5, fplus(va, vb))
        vg = fminus(va, vb)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            va, vb: `tuple` of `numpy.ndarray`
                The tie point vector representations of the tie point
                latitude-longitude representations.

            cv: `tuple` of `numpy.ndarray`
                The three-dimensional cartesian interpolation
                parameters.

        :Returns:

            `tuple` of `numpy.ndarray`

        """
        vr = self._fmultiply(0.5, self._fplus(va, vb))
        vg = self._fminus(va, vb)
        rsqr = self._fdot(vr, vr)
        gsqr = self._fdot(vg, vg)

        return (
            self._fdot(cv, self._fminus(va, vb)) / gsqr,
            self._fdot(cv, self._fcross(va, vb)) / (rsqr * gsqr),
        )

    def _fdot(self, va, vb):
        """Vector dot product.

         d = fdot(va, vb)
           = va.x*vb.x + va.y*vb.y + va.z*vb.z

         .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            va, vb: `tuple` of `numpy.ndarray`
                The vector representations of the tie point (latitude,
                longitude)

        :Returns:

            `numpy.ndarray`

        """
        return va[x] * vb[x] + va[y] * vb[y] + va[z] * vb[z]

    def _fll2v(self, lat, lon):
        """Vector representation of (latitude, longitude).

        Conversion from geocentric (latitude, longitude) to
        three-dimensional cartesian vector (x, y, z)

        (x, y, z) = fll2v(lat, lon)
                  = (cos(lat)*cos(lon),
                     cos(lat)*sin(lon),
                     sin(lat))

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            lat: `numpy.ndarray`
                Geocentric latitudes in degrees (not radians).

            lon: `numpy.ndarray`
                Geocentric longitudes in degrees (not radians).

        :Returns:

            `tuple` of `numpy.ndarray`

        """
        lat = np.deg2rad(lat)
        lon = np.deg2rad(lon)

        cos_lat = np.cos(lat)

        return (cos_lat * np.cos(lon), cos_lat * np.sin(lon), np.sin(lat))

    def _fminus(self, va, vb):
        """Vector difference.

        (x, y, z) = fminus(va, vb)
                  = (va.x - vb.x,
                     va.y - vb.y,
                     va.z - vb.z)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            va: `tuple` of `numpy.ndarray`

            vb: `tuple` of `numpy.ndarray`

        :Returns:

            `tuple` of `numpy.ndarray`

        """
        return tuple(a - b for a, b in zip(va, vb))

    def _fmultiply(self, r, v):
        """Vector multiplied by scalar.

        (x, y, z) = fmultiply(r, v)
                  = (r * v.x,
                     r * v.y,
                     r * v.z)

        .. versionaddedd:: (cfdm) 1.10.0.0

        :Parameters:

            r: scalar `numpy.ndarray`

            v: `tuple` of `numpy.ndarray`

        :Returns:

            `tuple` of `numpy.ndarray`

        """
        return tuple(a * r for a in v)

    def _fplus(self, *vectors):
        """Vector sum.

        Two vectors:

        (x, y, z) = fplus(va, vb)
                  = (va.x + vb.x,
                     va.y + vb.y,
                     va.z + vb.z)

        Three vectors:

        (x, y, z) = fplus(va, vb, vc)
                  = (va.x + vb.x + vc.x,
                     va.y + vb.y + vc.y,
                     va.z + vb.z + vc.z)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

           vectors:
               The vectors to be added, each defined by a `tuple` of
               `numpy.ndarray`.

        :Returns:

            `tuple` of `numpy.ndarray`

        """
        out = []
        for v in zip(*vectors):
            s = v[0]
            for a in v[1:]:
                s = s + a

            out.append(s)

        return tuple(out)

    def _fqv(self, va, vb, wv, d, s=None):
        """Quadratically interpolate 3-d cartesian coordinates.

        (x, y, z) = fqv(va, vb, wv, s)
                  = (fq(va.x, vb.x, wv.x, s),
                     fq(va.y, vb.y, wv.y, s),
                     fq(va.z, vb.z, wv.z, s))

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            va: `tuple` of `numpy.ndarray`
                The three-dimensionsal (x, y, z) vector representation
                of the first point along the subsampled dimension.

            vb: `tuple` of `numpy.ndarray`
                The three-dimensionsal (x, y, z) vector representation
                of the second point along the subsampled dimension.

            wv: `tuple` of `numpy.ndarray`
                The three-dimensional cartesian representation of the
                quadratic interpolation parameter ``w``.

            {{d: `int`}}

            {{s: array_like, optional}}

        :Returns:

            `tuple` of `numpy.ndarray`
                The three-dimensionsal (x, y, z) vector representation
                of the the interpolated points along the subsampled
                dimension.

        """
        return tuple(self._fq(va[i], vb[i], wv[i], d, s=s) for i in (x, y, z))

    def _fsqrt(self, t):
        """Square root.

        s = fsqrt(t)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            t: `numpy.ndarray`

        :Returns:

            `numpy.ndarray`
                The square root of the values.

        """
        return t**0.5

    def _fv2lat(self, v):
        """Convert cartesian to geocentric latitude coordinates.

        lat = fv2lat(v)
            = atan2(v.z, sqrt(v.x * v.x + v.y * v.y)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            v: `tuple` of `numpy.ndarray`
                The cartesian (x, y, z) coordinates.

        :Returns:

            `numpy.ndarray`
                Geocentric latitudes in degrees (not radians).

        """
        return np.rad2deg(
            np.arctan2(v[z], self._fsqrt(v[x] * v[x] + v[y] * v[y]))
        )

    def _fv2lon(self, v):
        """Convert cartesian to geocentric longitude coordinates.

        lon = fv2lon(v)
            = atan2(v.y, v.x)

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            v: `tuple` of `numpy.ndarray`
                The cartesian (x, y), or (x, y, z), coordinates.

        :Returns:

           `numpy.ndarray`
               Geocentric longitudes in degrees (not radians).

        """
        return np.rad2deg(np.arctan2(v[y], v[x]))
